# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from typing import Union

from pyforgejo.types import Label as PyforgejoLabel

from ogr.abstract import Issue, IssueLabel, Label, PRLabel, PullRequest


class ForgejoLabel(Label):
    def __init__(
        self,
        raw_label: PyforgejoLabel,
        parent: Union[PullRequest, Issue],
    ) -> None:
        super().__init__(parent)
        self._raw_label = raw_label

    def __str__(self) -> str:
        return f'ForgejoLabel(name="{self.name}")'

    @property
    def name(self):
        return self._raw_label.name


class ForgejoPRLabel(ForgejoLabel, PRLabel):
    pass


class ForgejoIssueLabel(ForgejoLabel, IssueLabel):
    pass



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from ogr.services.forgejo.issue import ForgejoIssue
from ogr.services.forgejo.project import ForgejoProject
from ogr.services.forgejo.pull_request import ForgejoPullRequest
from ogr.services.forgejo.service import ForgejoService

__all__ = [
    ForgejoPullRequest.__name__,
    ForgejoIssue.__name__,
    ForgejoProject.__name__,
    ForgejoService.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.access_level import AccessLevel
from ogr.abstract.auth_method import AuthMethod
from ogr.abstract.comment import (
    AnyComment,
    Comment,
    CommitComment,
    IssueComment,
    PRComment,
    Reaction,
)
from ogr.abstract.commit_flag import CommitFlag
from ogr.abstract.git_project import GitProject
from ogr.abstract.git_service import GitService
from ogr.abstract.git_tag import GitTag
from ogr.abstract.git_user import GitUser
from ogr.abstract.issue import Issue
from ogr.abstract.label import IssueLabel, Label, PRLabel
from ogr.abstract.pull_request import PullRequest
from ogr.abstract.release import Release
from ogr.abstract.status import CommitStatus, IssueStatus, MergeCommitStatus, PRStatus

__all__ = [
    OgrAbstractClass.__name__,
    AccessLevel.__name__,
    AuthMethod.__name__,
    AnyComment.__name__,
    Comment.__name__,
    CommitComment.__name__,
    IssueComment.__name__,
    PRComment.__name__,
    Reaction.__name__,
    CommitFlag.__name__,
    GitProject.__name__,
    GitService.__name__,
    GitTag.__name__,
    GitUser.__name__,
    Issue.__name__,
    IssueLabel.__name__,
    Label.__name__,
    PRLabel.__name__,
    PullRequest.__name__,
    Release.__name__,
    CommitStatus.__name__,
    IssueStatus.__name__,
    MergeCommitStatus.__name__,
    PRStatus.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from collections.abc import Iterable
from typing import Any, ClassVar, Optional, Union

from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.git_project import GitProject
from ogr.abstract.status import CommitStatus


class CommitFlag(OgrAbstractClass):
    _states: ClassVar[dict[str, CommitStatus]] = {}

    def __init__(
        self,
        raw_commit_flag: Optional[Any] = None,
        project: Optional["GitProject"] = None,
        commit: Optional[str] = None,
        state: Optional[CommitStatus] = None,
        context: Optional[str] = None,
        comment: Optional[str] = None,
        uid: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        self.uid = uid
        self.project = project
        self.commit = commit

        if commit and state and context:
            self.state = state
            self.context = context
            self.comment = comment
            self.url = url
        else:
            self._raw_commit_flag = raw_commit_flag
            self._from_raw_commit_flag()

    def __str__(self) -> str:
        return (
            f"CommitFlag("
            f"commit='{self.commit}', "
            f"state='{self.state.name}', "
            f"context='{self.context}', "
            f"uid='{self.uid}', "
            f"comment='{self.comment}', "
            f"url='{self.url}', "
            f"created='{self.created}', "
            f"edited='{self.edited}')"
        )

    @classmethod
    def _state_from_str(cls, state: str) -> CommitStatus:
        """
        Transforms state from string to enumeration.

        Args:
            state: String representation of a state.

        Returns:
            Commit status.
        """
        raise NotImplementedError()

    @classmethod
    def _validate_state(cls, state: CommitStatus) -> CommitStatus:
        """
        Validates state of the commit status (if it can be used with forge).
        """
        raise NotImplementedError()

    def _from_raw_commit_flag(self) -> None:
        """
        Sets attributes based on the raw flag that has been given through constructor.
        """
        raise NotImplementedError()

    @staticmethod
    def get(
        project: Any,
        commit: str,
    ) -> Union[list["CommitFlag"], Iterable["CommitFlag"]]:
        """
        Acquire commit statuses for given commit in the project.

        Args:
            project (GitProject): Project where the commit is located.
            commit: Commit hash for which we request statuses.

        Returns:
            List of commit statuses for the commit.
        """
        raise NotImplementedError()

    @staticmethod
    def set(
        project: Any,
        commit: str,
        state: CommitStatus,
        target_url: str,
        description: str,
        context: str,
    ) -> "CommitFlag":
        """
        Set a new commit status.

        Args:
            project (GitProject): Project where the commit is located.
            commit: Commit hash for which we set status.
            state: State for the commit status.
            target_url: URL for the commit status.
            description: Description of the commit status.
            context: Identifier to group related commit statuses.
        """
        raise NotImplementedError()

    @property
    def created(self) -> datetime.datetime:
        """Datetime of creating the commit status."""
        raise NotImplementedError()

    @property
    def edited(self) -> datetime.datetime:
        """Datetime of editing the commit status."""
        raise NotImplementedError()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT


from collections.abc import Iterable

from requre.helpers import record_httpx

from ogr.abstract import CommitStatus


@record_httpx()
def test_get_commit_statuses(project):
    flags = project.get_commit_statuses(
        commit="11b37d913374b14f8519d16c2a2cca3ebc14ac64",
    )

    assert isinstance(flags, Iterable)

    it = iter(flags)
    flag = next(it, None)

    assert flag.state == CommitStatus.success
    assert flag.comment == "testing status"
    assert flag.context == "test"
    assert flag.created.year == 2025


@record_httpx()
def test_set_commit_status(project):
    old_statuses = project.get_commit_statuses(
        commit="11b37d913374b14f8519d16c2a2cca3ebc14ac64",
    )

    status = project.set_commit_status(
        commit="11b37d913374b14f8519d16c2a2cca3ebc14ac64",
        state=CommitStatus.success,
        target_url="https://v10.next.forgejo.org/packit-validator/ogr-tests",
        description="testing status",
        context="test",
    )

    assert status
    new_statuses = project.get_commit_statuses(
        commit="11b37d913374b14f8519d16c2a2cca3ebc14ac64",
    )

    old_count = sum(1 for _ in iter(old_statuses))
    new_count = sum(1 for _ in iter(new_statuses))

    assert old_count == new_count



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from collections.abc import Iterable
from re import Match
from typing import Any, Optional, Union

from ogr import abstract as _abstract
from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.commit_flag import CommitFlag
from ogr.abstract.git_project import GitProject
from ogr.abstract.status import MergeCommitStatus, PRStatus


class PullRequest(OgrAbstractClass):
    """
    Attributes:
        project (GitProject): Project of the pull request.
    """

    def __init__(self, raw_pr: Any, project: "GitProject") -> None:
        self._raw_pr = raw_pr
        self._target_project = project

    @property
    def title(self) -> str:
        """Title of the pull request."""
        raise NotImplementedError()

    @title.setter
    def title(self, new_title: str) -> None:
        raise NotImplementedError()

    @property
    def id(self) -> int:
        """ID of the pull request."""
        raise NotImplementedError()

    @property
    def status(self) -> PRStatus:
        """Status of the pull request."""
        raise NotImplementedError()

    @property
    def url(self) -> str:
        """Web URL of the pull request."""
        raise NotImplementedError()

    @property
    def description(self) -> str:
        """Description of the pull request."""
        raise NotImplementedError()

    @description.setter
    def description(self, new_description: str) -> None:
        raise NotImplementedError

    @property
    def author(self) -> str:
        """Login of the author of the pull request."""
        raise NotImplementedError()

    @property
    def source_branch(self) -> str:
        """Name of the source branch (from which the changes are pulled)."""
        raise NotImplementedError()

    @property
    def target_branch(self) -> str:
        """Name of the target branch (where the changes are being merged)."""
        raise NotImplementedError()

    @property
    def created(self) -> datetime.datetime:
        """Datetime of creating the pull request."""
        raise NotImplementedError()

    @property
    def labels(self) -> Union[list["_abstract.PRLabel"], Iterable["_abstract.PRLabel"]]:
        """Labels of the pull request."""
        raise NotImplementedError()

    @property
    def diff_url(self) -> str:
        """Web URL to the diff of the pull request."""
        raise NotImplementedError()

    @property
    def patch(self) -> bytes:
        """Patch of the pull request."""
        raise NotImplementedError()

    @property
    def head_commit(self) -> str:
        """Commit hash of the HEAD commit of the pull request."""
        raise NotImplementedError()

    @property
    def target_branch_head_commit(self) -> str:
        """Commit hash of the HEAD commit of the target branch."""
        raise NotImplementedError()

    @property
    def merge_commit_sha(self) -> str:
        """
        Commit hash of the merge commit of the pull request.

        Before merging represents test merge commit, if git forge supports it.
        """
        raise NotImplementedError()

    @property
    def merge_commit_status(self) -> MergeCommitStatus:
        """Current status of the test merge commit."""
        raise NotImplementedError()

    @property
    def source_project(self) -> "GitProject":
        """Object that represents source project (from which the changes are pulled)."""
        raise NotImplementedError()

    @property
    def target_project(self) -> "GitProject":
        """Object that represents target project (where changes are merged)."""
        return self._target_project

    @property
    def commits_url(self) -> str:
        """Web URL to the list of commits in the pull request."""
        raise NotImplementedError()

    @property
    def closed_by(self) -> Optional[str]:
        """Login of the account that closed the pull request."""
        raise NotImplementedError()

    def __str__(self) -> str:
        description = (
            f"{self.description[:10]}..." if self.description is not None else "None"
        )
        return (
            f"PullRequest("
            f"title='{self.title}', "
            f"id={self.id}, "
            f"status='{self.status.name}', "
            f"url='{self.url}', "
            f"diff_url='{self.diff_url}', "
            f"description='{description}', "
            f"author='{self.author}', "
            f"source_branch='{self.source_branch}', "
            f"target_branch='{self.target_branch}', "
            f"created='{self.created}')"
        )

    @staticmethod
    def create(
        project: Any,
        title: str,
        body: str,
        target_branch: str,
        source_branch: str,
        fork_username: Optional[str] = None,
    ) -> "PullRequest":
        """
        Create new pull request.

        Args:
            project (GitProject): Project where the pull request will be created.
            title: Title of the pull request.
            body: Description of the pull request.
            target_branch: Branch in the project where the changes are being
                merged.
            source_branch: Branch from which the changes are being pulled.
            fork_username: The username/namespace of the forked repository.

        Returns:
            Object that represents newly created pull request.
        """
        raise NotImplementedError()

    @staticmethod
    def get(project: Any, id: int) -> "PullRequest":
        """
        Get pull request.

        Args:
            project (GitProject): Project where the pull request is located.
            id: ID of the pull request.

        Returns:
            Object that represents pull request.
        """
        raise NotImplementedError()

    @staticmethod
    def get_list(
        project: Any,
        status: PRStatus = PRStatus.open,
    ) -> Union[list["PullRequest"], Iterable["PullRequest"]]:
        """
        List of pull requests.

        Args:
            project (GitProject): Project where the pull requests are located.
            status: Filters out the pull requests.

                Defaults to `PRStatus.open`.

        Returns:
            List of pull requests with requested status.
        """
        raise NotImplementedError()

    def update_info(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "PullRequest":
        """
        Update pull request information.

        Args:
            title: The new title of the pull request.

                Defaults to `None`, which means no updating.
            description: The new description of the pull request.

                Defaults to `None`, which means no updating.

        Returns:
            Pull request itself.
        """
        raise NotImplementedError()

    def _get_all_comments(
        self,
        reverse: bool = False,
    ) -> Union[list["_abstract.PRComment"], Iterable["_abstract.PRComment"]]:
        """
        Get list of all pull request comments.

        Args:
            reverse: Defines whether the comments should be listed in a reversed
                order.

                Defaults to `False`.

        Returns:
            List of all comments on the pull request.
        """
        raise NotImplementedError()

    def get_comments(
        self,
        filter_regex: Optional[str] = None,
        reverse: bool = False,
        author: Optional[str] = None,
    ) -> Union[list["_abstract.PRComment"], Iterable["_abstract.PRComment"]]:
        """
        Get list of pull request comments.

        Args:
            filter_regex: Filter the comments' content with `re.search`.

                Defaults to `None`, which means no filtering.
            reverse: Whether the comments are to be returned in
                reversed order.

                Defaults to `False`.
            author: Filter the comments by author.

                Defaults to `None`, which means no filtering.

        Returns:
            List of pull request comments.
        """
        raise NotImplementedError()

    def get_all_commits(self) -> Union[list[str], Iterable[str]]:
        """
        Returns:
            List of commit hashes of commits in pull request.
        """
        raise NotImplementedError()

    def search(
        self,
        filter_regex: str,
        reverse: bool = False,
        description: bool = True,
    ) -> Optional[Match[str]]:
        """
        Find match in pull request description or comments.

        Args:
            filter_regex: Regex that is used to filter the comments' content with `re.search`.
            reverse: Reverse order of the comments.

                Defaults to `False`.
            description: Whether description is included in the search.

                Defaults to `True`.

        Returns:
            `re.Match` if found, `None` otherwise.
        """
        raise NotImplementedError()

    def comment(
        self,
        body: str,
        commit: Optional[str] = None,
        filename: Optional[str] = None,
        row: Optional[int] = None,
    ) -> "_abstract.PRComment":
        """
        Add new comment to the pull request.

        Args:
            body: Body of the comment.
            commit: Commit hash to which comment is related.

                Defaults to generic comment.
            filename: Path to the file to which comment is related.

                Defaults to no relation to the file.
            row: Line number to which the comment is related.

                Defaults to no relation to the line.

        Returns:
            Newly created comment.
        """
        raise NotImplementedError()

    def close(self) -> "PullRequest":
        """
        Close the pull request.

        Returns:
            Pull request itself.
        """
        raise NotImplementedError()

    def merge(self) -> "PullRequest":
        """
        Merge the pull request.

        Returns:
            Pull request itself.
        """
        raise NotImplementedError()

    def add_label(self, *labels: str) -> None:
        """
        Add labels to the pull request.

        Args:
            *labels: Labels to be added.
        """
        raise NotImplementedError()

    def get_statuses(self) -> Union[list["CommitFlag"], Iterable["CommitFlag"]]:
        """
        Returns statuses for latest commit on pull request.

        Returns:
            List of commit statuses of the latest commit.
        """
        raise NotImplementedError()

    def get_comment(self, comment_id: int) -> "_abstract.PRComment":
        """
        Returns a PR comment.

        Args:
            comment_id: id of comment

        Returns:
            Object representing a PR comment.
        """
        raise NotImplementedError()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from collections.abc import Iterable, Sequence
from typing import Union

from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.git_project import GitProject
from ogr.abstract.git_service import GitService


class GitUser(OgrAbstractClass):
    """
    Represents currently authenticated user through service.
    """

    def __init__(self, service: GitService) -> None:
        self.service = service

    def get_username(self) -> str:
        """
        Returns:
            Login of the user.
        """
        raise NotImplementedError()

    def get_email(self) -> str:
        """
        Returns:
            Email of the user.
        """
        raise NotImplementedError()

    def get_projects(self) -> Union[Sequence["GitProject"], Iterable["GitProject"]]:
        """
        Returns:
            Sequence of projects in user's namespace.
        """
        raise NotImplementedError()

    def get_forks(self) -> Union[Sequence["GitProject"], Iterable["GitProject"]]:
        """
        Returns:
            Sequence of forks in user's namespace.
        """
        raise NotImplementedError()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from enum import Enum


class AuthMethod(str, Enum):
    tokman = "tokman"
    github_app = "github_app"
    token = "token"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from collections.abc import Iterable
from typing import Any, Optional, Union

from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.git_project import GitProject
from ogr.abstract.git_tag import GitTag


class Release(OgrAbstractClass):
    """
    Object that represents release.

    Attributes:
        project (GitProject): Project on which the release is created.
    """

    def __init__(
        self,
        raw_release: Any,
        project: "GitProject",
    ) -> None:
        self._raw_release = raw_release
        self.project = project

    def __str__(self) -> str:
        return (
            f"Release("
            f"title='{self.title}', "
            f"body='{self.body}', "
            f"tag_name='{self.tag_name}', "
            f"url='{self.url}', "
            f"created_at='{self.created_at}', "
            f"tarball_url='{self.tarball_url}')"
        )

    @property
    def title(self) -> str:
        """Title of the release."""
        raise NotImplementedError()

    @property
    def body(self) -> str:
        """Body of the release."""
        raise NotImplementedError()

    @property
    def git_tag(self) -> GitTag:
        """Object that represents tag tied to the release."""
        raise NotImplementedError()

    @property
    def tag_name(self) -> str:
        """Tag tied to the release."""
        raise NotImplementedError()

    @property
    def url(self) -> Optional[str]:
        """URL of the release."""
        raise NotImplementedError()

    # TODO: Check if should really be string
    @property
    def created_at(self) -> datetime.datetime:
        """Datetime of creating the release."""
        raise NotImplementedError()

    @property
    def tarball_url(self) -> str:
        """URL of the tarball."""
        raise NotImplementedError()

    @staticmethod
    def get(
        project: Any,
        identifier: Optional[int] = None,
        name: Optional[str] = None,
        tag_name: Optional[str] = None,
    ) -> "Release":
        """
        Get a single release.

        Args:
            identifier: Identifier of the release.

                Defaults to `None`, which means not being used.
            name: Name of the release.

                Defaults to `None`, which means not being used.
            tag_name: Tag that the release is tied to.

                Defaults to `None`, which means not being used.

        Returns:
            Object that represents release that satisfies requested condition.
        """
        raise NotImplementedError()

    @staticmethod
    def get_latest(project: Any) -> Optional["Release"]:
        """
        Returns:
            Object that represents the latest release.
        """
        raise NotImplementedError()

    @staticmethod
    def get_list(project: Any) -> Union[list["Release"], Iterable["Release"]]:
        """
        Returns:
            List of the objects that represent releases.
        """
        raise NotImplementedError()

    @staticmethod
    def create(
        project: Any,
        tag: str,
        name: str,
        message: str,
        ref: Optional[str] = None,
    ) -> "Release":
        """
        Create new release.

        Args:
            project: Project where the release is to be created.
            tag: Tag which is the release based off.
            name: Name of the release.
            message: Message or description of the release.
            ref: Git reference, mainly commit hash for the release. If provided
                git tag is created prior to creating a release.

                Defaults to `None`.

        Returns:
            Object that represents newly created release.
        """
        raise NotImplementedError()

    def save_archive(self, filename: str) -> None:
        """
        Save tarball of the release to requested `filename`.

        Args:
            filename: Path to the file to save archive to.
        """
        raise NotImplementedError()

    def edit_release(self, name: str, message: str) -> None:
        """
        Edit name and message of a release.

        Args:
            name: Name of the release.
            message: Description of the release.
        """
        raise NotImplementedError()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pyforgejo
import pytest
from requre.helpers import record_httpx


@pytest.mark.parametrize(
    "kwargs_",
    [
        pytest.param(
            {"repo": "test", "namespace": "packit"},
            id="create project with specified namespace (organization)",
        ),
        pytest.param(
            {"repo": "test", "namespace": None},
            id="create project without namespace (in the user's namespace)",
        ),
        pytest.param(
            {"repo": "test_1", "namespace": None, "description": "A repo description"},
            id="create project with description",
        ),
    ],
)
@record_httpx()
def test_project_create(service, kwargs_):
    # Construct params for fetching the project
    kwargs_fetch = kwargs_.copy()
    kwargs_fetch["namespace"] = kwargs_fetch["namespace"] or service.user.get_username()

    # Check that project doesn't exist
    project = service.get_project(**kwargs_fetch)
    with pytest.raises(pyforgejo.core.api_error.ApiError):
        _ = project.forgejo_repo

    # Create new project
    new_project = service.project_create(**kwargs_)
    assert new_project.repo == kwargs_["repo"]
    assert new_project.forgejo_repo

    # Try to fetch newly created project
    project = service.get_project(**kwargs_fetch)
    assert project.forgejo_repo



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import functools
from typing import Any, Callable, Union

import github
import gitlab
import pyforgejo
import requests

from ogr.exceptions import (
    APIException,
    ForgejoAPIException,
    GitForgeInternalError,
    GithubAPIException,
    GitlabAPIException,
    OgrNetworkError,
)


def __check_for_internal_failure(ex: APIException):
    """
    Checks if exception is caused by internal failure from git forge.

    Args:
        ex: Wrapped exception.

    Raises:
        GitForgeInternalError, when exception was cause by an internal failure.
        APIException, exception itself when not an internal failure.
    """
    if ex.response_code is not None and ex.response_code >= 500:
        raise GitForgeInternalError from ex.__cause__
    raise ex


def __wrap_exception(
    ex: Union[
        github.GithubException,
        gitlab.GitlabError,
        pyforgejo.core.api_error.ApiError,  # type: ignore
    ],
) -> APIException:
    """
    Wraps uncaught exception in one of ogr exceptions.

    Args:
        ex: Unhandled exception from GitHub or GitLab.

    Returns:
        Wrapped `ex` in respective `APIException`.

    Raises:
        TypeError, when given unexpected type of exception.
    """
    MAPPING = {
        github.GithubException: GithubAPIException,
        gitlab.GitlabError: GitlabAPIException,
        pyforgejo.core.api_error.ApiError: ForgejoAPIException,  # type: ignore
    }

    for caught_exception, ogr_exception in MAPPING.items():
        if isinstance(ex, caught_exception):
            exc = ogr_exception(str(ex))
            exc.__cause__ = ex
            return exc

    raise TypeError("Unknown type of uncaught exception passed") from ex


def catch_common_exceptions(function: Callable) -> Any:
    """
    Decorator catching common exceptions.

    Args:
        function (Callable): Function or method to decorate.

    Raises:
        GithubAPIException, if authentication to Github failed.
        GitlabAPIException, if authentication to Gitlab failed.
        OgrNetworkError, if network problems occurred while performing a request.
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except github.BadCredentialsException as ex:
            raise GithubAPIException("Invalid Github credentials") from ex
        except gitlab.GitlabAuthenticationError as ex:
            raise GitlabAPIException("Invalid Gitlab credentials") from ex
        except requests.exceptions.ConnectionError as ex:
            raise OgrNetworkError(
                "Could not perform the request due to a network error",
            ) from ex
        except APIException as ex:
            __check_for_internal_failure(ex)
        except (
            github.GithubException,
            gitlab.GitlabError,
            pyforgejo.core.api_error.ApiError,  # type: ignore
        ) as ex:
            __check_for_internal_failure(__wrap_exception(ex))

    return wrapper


class CatchCommonErrors(type):
    """
    A metaclass wrapping methods with a common exception handler.

    This handler catches exceptions which can occur almost anywhere
    and catching them manually would be tedious and converts them
    to an appropriate ogr exception for the user. This includes
    exceptions such as:
        - authentication (from Github/Gitlab)
        - network errors
    """

    def __new__(cls, name, bases, namespace):
        for key, value in namespace.items():
            # There is an anticipated change in behaviour in Python 3.10
            # for static/class methods. From Python 3.10 they will be callable.
            # We need to achieve consistent behaviour with older versions,
            # hence the explicit handling is needed here (isinstance checking
            # works the same). Moreover, static/class method decorator must
            # be used last, especially prior to Python 3.10 since they return
            # descriptor objects and not functions.
            # See: https://bugs.python.org/issue43682
            if isinstance(value, staticmethod):
                namespace[key] = staticmethod(catch_common_exceptions(value.__func__))
            elif isinstance(value, classmethod):
                namespace[key] = classmethod(catch_common_exceptions(value.__func__))
            elif callable(namespace[key]):
                namespace[key] = catch_common_exceptions(namespace[key])
        return super().__new__(cls, name, bases, namespace)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from functools import cached_property
from typing import Optional
from urllib.parse import urlparse

from pyforgejo import PyforgejoApi

from ogr.abstract import GitUser
from ogr.exceptions import OgrException
from ogr.factory import use_for_service
from ogr.services.base import BaseGitService
from ogr.services.forgejo.project import ForgejoProject
from ogr.services.forgejo.user import ForgejoUser


@use_for_service("forgejo")
@use_for_service("codeberg.org")
class ForgejoService(BaseGitService):
    version = "/api/v1"

    def __init__(
        self,
        instance_url: str = "https://codeberg.org",
        api_key: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()
        self.instance_url = instance_url + self.version
        self._token = f"token {api_key}"
        self._api = None

    @cached_property
    def api(self) -> PyforgejoApi:
        return PyforgejoApi(base_url=self.instance_url, api_key=self._token)

    def get_project(  # type: ignore[override]
        self,
        repo: str,
        namespace: str,
        **kwargs,
    ) -> "ForgejoProject":
        return ForgejoProject(
            repo=repo,
            namespace=namespace,
            service=self,
            **kwargs,
        )

    @property
    def user(self) -> GitUser:
        return ForgejoUser(self)

    def project_create(
        self,
        repo: str,
        namespace: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "ForgejoProject":
        if namespace:
            new_repo = self.api.organization.create_org_repo(
                org=namespace,
                name=repo,
                description=description,
            )
        else:
            new_repo = self.api.repository.create_current_user_repo(
                name=repo,
                description=description,
            )
        return ForgejoProject(
            repo=repo,
            namespace=namespace or self.user.get_username(),
            service=self,
            github_repo=new_repo,
        )

    def get_project_from_url(self, url: str) -> "ForgejoProject":
        parsed_url = urlparse(url)
        path_parts = parsed_url.path.strip("/").split("/")

        if len(path_parts) < 2:
            raise OgrException(f"Invalid Forgejo URL: {url}")

        namespace = path_parts[0]
        repo = path_parts[1]

        return self.get_project(repo=repo, namespace=namespace)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import datetime
from functools import cached_property, partial
from typing import Optional

from pyforgejo.types import Release as PyforgejoRelease

from ogr.abstract import GitTag, Release
from ogr.exceptions import ForgejoAPIException
from ogr.services import forgejo as ogr_forgejo
from ogr.services.forgejo.utils import paginate


class ForgejoRelease(Release):
    _raw_release: PyforgejoRelease
    project: "ogr_forgejo.ForgejoProject"

    @property
    def title(self) -> str:
        return self._raw_release.name

    @property
    def body(self) -> str:
        return self._raw_release.body

    @cached_property
    def git_tag(self) -> GitTag:
        tag = self.project.api.repo_get_tag(
            owner=self.project.namespace,
            repo=self.project.repo,
            tag=self.tag_name,
        )
        return GitTag(name=tag.name, commit_sha=tag.commit.sha)

    @property
    def tag_name(self) -> str:
        return self._raw_release.tag_name

    @property
    def url(self) -> Optional[str]:
        return self._raw_release.url

    @property
    def created_at(self) -> datetime.datetime:
        return self._raw_release.created_at

    @property
    def tarball_url(self) -> str:
        return self._raw_release.tarball_url

    @staticmethod
    def _release_id_from_name(
        project: "ogr_forgejo.ForgejoProject",
        name: str,
    ) -> Optional[int]:
        for release in paginate(
            partial(
                project.api.repo_list_releases,
                owner=project.namespace,
                repo=project.repo,
            ),
        ):
            if release.name == name:
                return release.id
        return None

    @staticmethod
    def get(
        project: "ogr_forgejo.ForgejoProject",
        identifier: Optional[int] = None,
        name: Optional[str] = None,
        tag_name: Optional[str] = None,
    ) -> "Release":
        if tag_name:
            release = project.api.repo_get_release_by_tag(
                owner=project.namespace,
                repo=project.repo,
                tag=tag_name,
            )
            return ForgejoRelease(release, project)

        if name:
            identifier = ForgejoRelease._release_id_from_name(project, name)

        if identifier is None:
            raise ForgejoAPIException("Release was not found.")

        release = project.api.repo_get_release(
            owner=project.namespace,
            repo=project.repo,
            id=identifier,
        )
        return ForgejoRelease(release, project)

    @staticmethod
    def get_latest(project: "ogr_forgejo.ForgejoProject") -> Optional["Release"]:
        releases = project.api.repo_list_releases(
            owner=project.namespace,
            repo=project.repo,
            page=1,
            limit=1,
        )

        return ForgejoRelease(releases[0], project) if releases else None

    @staticmethod
    def get_list(project: "ogr_forgejo.ForgejoProject") -> list["Release"]:
        releases = paginate(
            partial(
                project.api.repo_list_releases,
                owner=project.namespace,
                repo=project.repo,
            ),
        )
        return [ForgejoRelease(release, project) for release in releases]

    @staticmethod
    def create(
        project: "ogr_forgejo.ForgejoProject",
        tag: str,
        name: str,
        message: str,
        ref: Optional[str] = None,
    ) -> "Release":
        release = project.api.repo_create_release(
            owner=project.namespace,
            repo=project.repo,
            tag_name=tag,
            body=message,
            name=name,
            target_commitish=ref,
        )

        return ForgejoRelease(release, project)

    def edit_release(self, name: str, message: str) -> None:
        try:
            data = {}
            if name is not None:
                data["name"] = name

            if message is not None:
                data["body"] = message

            updated_release = self.project.api.repo_edit_release(
                owner=self.project.namespace,
                repo=self.project.repo,
                id=self._raw_release.id,
                **data,
            )

            self._raw_release = updated_release
        except Exception as ex:
            raise ForgejoAPIException(
                f"There was an error while updating Forgejo release: {ex}",
            ) from ex



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from collections.abc import Iterable
from typing import Any, Optional, TypeVar, Union

from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.issue import Issue
from ogr.abstract.pull_request import PullRequest
from ogr.deprecation import deprecate_and_set_removal

AnyComment = TypeVar("AnyComment", bound="Comment")


class Reaction(OgrAbstractClass):
    def __init__(self, raw_reaction: Any) -> None:
        self._raw_reaction = raw_reaction

    def __str__(self):
        return f"Reaction(raw_reaction={self._raw_reaction})"

    def delete(self) -> None:
        """Delete a reaction."""
        raise NotImplementedError()


class Comment(OgrAbstractClass):
    def __init__(
        self,
        raw_comment: Optional[Any] = None,
        parent: Optional[Any] = None,
        body: Optional[str] = None,
        id_: Optional[int] = None,
        author: Optional[str] = None,
        created: Optional[datetime.datetime] = None,
        edited: Optional[datetime.datetime] = None,
    ) -> None:
        if raw_comment:
            self._from_raw_comment(raw_comment)
        elif body and author:
            self._body = body
            self._id = id_
            self._author = author
            self._created = created
            self._edited = edited
        else:
            raise ValueError("cannot construct comment without body and author")

        self._parent = parent

    def __str__(self) -> str:
        body = f"{self.body[:10]}..." if self.body is not None else "None"
        return (
            f"Comment("
            f"comment='{body}', "
            f"author='{self.author}', "
            f"created='{self.created}', "
            f"edited='{self.edited}')"
        )

    def _from_raw_comment(self, raw_comment: Any) -> None:
        """Constructs Comment object from raw_comment given from API."""
        raise NotImplementedError()

    @property
    def body(self) -> str:
        """Body of the comment."""
        return self._body

    @body.setter
    def body(self, new_body: str) -> None:
        self._body = new_body

    @property
    def id(self) -> int:
        return self._id

    @property
    def author(self) -> str:
        """Login of the author of the comment."""
        return self._author

    @property
    def created(self) -> datetime.datetime:
        """Datetime of creation of the comment."""
        return self._created

    @property
    def edited(self) -> datetime.datetime:
        """Datetime of last edit of the comment."""
        return self._edited

    def get_reactions(self) -> Union[list[Reaction], Iterable[Reaction]]:
        """Returns list of reactions."""
        raise NotImplementedError()

    def add_reaction(self, reaction: str) -> Reaction:
        """
        Reacts to a comment.

        Colons in between reaction are not needed, e.g. `comment.add_reaction("+1")`.

        Args:
            reaction: String representing specific reaction to be added.

        Returns:
            Object representing newly added reaction.
        """
        raise NotImplementedError()


class IssueComment(Comment):
    @property
    def issue(self) -> "Issue":
        """Issue of issue comment."""
        return self._parent

    def __str__(self) -> str:
        return "Issue" + super().__str__()


class CommitComment(Comment):
    """
    Attributes:
        sha (str): Hash of the related commit.
        body (str): Body of the comment.
        author (str): Login of the author.
    """

    def __init__(
        self,
        sha: str,
        raw_comment: Any,
    ) -> None:
        super().__init__(raw_comment=raw_comment)
        self.sha = sha

    @property  # type: ignore
    @deprecate_and_set_removal(
        since="0.41.0",
        remove_in="0.46.0 (or 1.0.0 if it comes sooner)",
        message="Use body",
    )
    def comment(self) -> str:
        return self.body

    def __str__(self) -> str:
        return (
            f"CommitComment(commit={self.sha}, author={self.author}, body={self.body})"
        )


class PRComment(Comment):
    @property
    def pull_request(self) -> "PullRequest":
        """Pull request of pull request comment."""
        return self._parent

    def __str__(self) -> str:
        return "PR" + super().__str__()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT


from functools import cached_property

from ogr.services import forgejo
from ogr.services.base import BaseGitUser
from ogr.services.forgejo.project import ForgejoProject


class ForgejoUser(BaseGitUser):
    service: "forgejo.ForgejoService"

    def __init__(self, service: "forgejo.ForgejoService") -> None:
        super().__init__(service=service)
        self._forgejo_user = None

    def __str__(self) -> str:
        return f'ForgejoUser(username="{self.get_username()}")'

    @cached_property
    def forgejo_user(self):
        return self.service.api.user.get_current()

    def get_username(self) -> str:
        return self.forgejo_user.login

    def get_email(self) -> str:
        return self.forgejo_user.email

    def get_projects(self) -> list["ForgejoProject"]:
        repos = self.service.api.user.current_list_repos()
        return [
            ForgejoProject(
                repo=repo.name,
                namespace=repo.owner.login,
                service=self.service,
                forgejo_repo=repo,
            )
            for repo in repos
        ]

    def get_forks(self) -> list["ForgejoProject"]:
        return [project for project in self.get_projects() if project.forgejo_repo.fork]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
import logging
from collections.abc import Iterable
from functools import partial
from typing import ClassVar, Union

import pyforgejo

from ogr.abstract import CommitStatus
from ogr.exceptions import ForgejoAPIException
from ogr.services import forgejo as ogr_forgejo
from ogr.services.base import BaseCommitFlag

from .utils import paginate

logger = logging.getLogger(__name__)


class ForgejoCommitFlag(BaseCommitFlag):
    _states: ClassVar[dict[str, CommitStatus]] = {
        "pending": CommitStatus.pending,
        "success": CommitStatus.success,
        "failure": CommitStatus.failure,
        "canceled": CommitStatus.error,
        "warning": CommitStatus.warning,
    }

    @staticmethod
    def _state_from_enum(state: CommitStatus) -> str:
        return state.name

    def __str__(self) -> str:
        return (
            f"ForgejoCommitFlag("
            f"state='{self.state.name}', "
            f"context='{self.context}', "
            f"uid='{self.uid}', "
            f"comment='{self.comment}', "
            f"url='{self.url}', "
            f"created='{self.created}')"
        )

    def _from_raw_commit_flag(self):
        self.state = self._state_from_str(self._raw_commit_flag.status)
        self.context = self._raw_commit_flag.context
        self.comment = self._raw_commit_flag.description
        self.uid = self._raw_commit_flag.id
        self.url = self._raw_commit_flag.target_url

    @staticmethod
    def get(
        project: "ogr_forgejo.ForgejoProject",
        commit: str,
    ) -> Iterable["ForgejoCommitFlag"]:

        try:
            return (
                ForgejoCommitFlag(
                    raw_commit_flag=status,
                    project=project,
                    commit=commit,
                )
                for status in paginate(
                    partial(
                        project.api.repo_list_statuses,
                        owner=project.namespace,
                        repo=project.repo,
                        sha=commit,
                    ),
                )
            )
        except pyforgejo.NotFoundError as ex:  # 404 error
            logger.error(f"Commit {commit} was not found.")
            raise ForgejoAPIException(f"Commit {commit} was not found.") from ex

    @staticmethod
    def set(
        project: "ogr_forgejo.ForgejoProject",
        commit: str,
        state: Union[CommitStatus, str],
        target_url: str,
        description: str,
        context: str,
        trim: bool = False,
    ) -> "ForgejoCommitFlag":

        if isinstance(state, str):
            state = ForgejoCommitFlag._state_from_str(state)

        state = ForgejoCommitFlag._validate_state(state)

        if trim:
            description = description[:140]

        owner = project.forgejo_repo.owner.login
        forgejo_repo = project.forgejo_repo

        try:
            status = project.api.repo_create_status(
                owner=owner,
                repo=forgejo_repo.name,
                sha=commit,
                context=context,
                description=description,
                state=ForgejoCommitFlag._state_from_enum(state),
                target_url=target_url,
                request_options=None,
            )

        except pyforgejo.NotFoundError as ex:  # 404 error
            logger.error(f"Commit {commit} was not found.")
            raise ForgejoAPIException(f"Commit {commit} was not found.") from ex

        return ForgejoCommitFlag(project=project, raw_commit_flag=status, commit=commit)

    @property
    def created(self) -> datetime.datetime:
        return self._raw_commit_flag.created_at

    @property
    def edited(self) -> datetime.datetime:
        return self._raw_commit_flag.updated_at



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from ogr.abstract.exception import CatchCommonErrors


class OgrAbstractClass(metaclass=CatchCommonErrors):
    def __repr__(self) -> str:
        return f"<{self!s}>"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from functools import partial


def paginate(api_call, /, *args, **kwargs):
    api_call = partial(api_call, *args, **kwargs)

    page = 1
    while objects := api_call(page=page):
        yield from objects

        page += 1



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from enum import Enum, IntEnum


class IssueStatus(IntEnum):
    """Enumeration for issue statuses."""

    open = 1
    closed = 2
    all = 3


class PRStatus(IntEnum):
    """Enumeration that represents statuses of pull requests."""

    open = 1
    closed = 2
    merged = 3
    all = 4


class CommitStatus(Enum):
    """Enumeration that represents possible state of commit statuses."""

    pending = 1
    success = 2
    failure = 3
    error = 4
    canceled = 5
    running = 6
    warning = 7


class MergeCommitStatus(Enum):
    """Enumeration that represents possible states of merge states of PR/MR."""

    can_be_merged = 1
    cannot_be_merged = 2
    unchecked = 3
    checking = 4
    cannot_be_merged_recheck = 5



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os

import pytest

from ogr.services.forgejo import ForgejoService


@pytest.fixture
def service():
    api_key = os.environ.get("FORGEJO_TOKEN")
    return ForgejoService(
        instance_url="https://v10.next.forgejo.org",
        api_key=api_key,
    )


@pytest.fixture
def project(service):
    return service.get_project(
        namespace="packit-validator",
        repo="ogr-tests",
    )


@pytest.fixture
def hello_world_project(service):
    return service.get_project(
        namespace="packit-validator",
        repo="hello-world",
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from collections.abc import Iterable, Sequence
from typing import Optional, Union

from ogr import abstract as _abstract
from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.access_level import AccessLevel
from ogr.abstract.git_tag import GitTag
from ogr.abstract.status import CommitStatus, IssueStatus, PRStatus


class GitProject(OgrAbstractClass):
    def __init__(
        self,
        repo: str,
        service: "_abstract.GitService",
        namespace: str,
    ) -> None:
        """
        Args:
            repo: Name of the project.
            service: GitService instance.
            namespace: Namespace of the project.

                - GitHub: username or org name.
                - GitLab: username or org name.
                - Pagure: namespace (e.g. `"rpms"`).

                  In case of forks: `"fork/{username}/{namespace}"`.
        """
        self.service = service
        self.repo = repo
        self.namespace = namespace

    def __str__(self) -> str:
        return f"GitProject(namespace={self.namespace}, repo={self.repo}, service={self.service})"

    @property
    def description(self) -> str:
        """
        Returns:
            Project description.
        """
        raise NotImplementedError()

    @description.setter
    def description(self, new_description: str) -> None:
        """
        Args:
            new_description: description to set for project.
        """
        raise NotImplementedError()

    def delete(self) -> None:
        """Delete the project."""
        raise NotImplementedError()

    def exists(self) -> bool:
        """
        Check the existence of the repo.

        Returns:
            `True` if the project exists, `False` otherwise.
        """
        raise NotImplementedError()

    def is_private(self) -> bool:
        """
        Is this repository private (accessible only by users with permissions).

        Returns:
            `True`, if the repository is private.
        """
        raise NotImplementedError()

    def is_forked(self) -> bool:
        """
        Is this repository forked by the authenticated user?

        Returns:
            `True`, if the repository is fork.
        """
        raise NotImplementedError()

    @property
    def is_fork(self) -> bool:
        """`True` if the project is a fork."""
        raise NotImplementedError()

    @property
    def full_repo_name(self) -> str:
        """Get repo name with namespace, e.g. `rpms/python-docker-py`."""
        raise NotImplementedError()

    @property
    def parent(self) -> Optional["GitProject"]:
        """Parent project if the project is a fork, otherwise `None`."""
        raise NotImplementedError()

    @property
    def has_issues(self) -> bool:
        """`True` if issues are enabled on the project."""
        raise NotImplementedError()

    def get_branches(self) -> Union[list[str], Iterable[str]]:
        """
        Returns:
            List with names of branches in the project.
        """
        raise NotImplementedError()

    @property
    def default_branch(self) -> str:
        """Default branch (usually `main`, `master` or `trunk`)."""
        raise NotImplementedError()

    def get_commits(self, ref: Optional[str] = None) -> Union[list[str], Iterable[str]]:
        """
        Get list of commits for the project.

        Args:
            ref: Ref to start listing commits from, defaults to the default project branch.

        Returns:
            List of commit SHAs for the project.
        """
        raise NotImplementedError()

    def get_description(self) -> str:
        """
        Returns:
            Project description.
        """
        raise NotImplementedError()

    def get_fork(self, create: bool = True) -> Optional["GitProject"]:
        """
        Provide GitProject instance of a fork of this project.

        Args:
            create: Create fork if it does not exist.

        Returns:
            `None` if the project is fork itself or there is no fork, otherwise
            instance of a fork if is to be created or exists already.
        """
        raise NotImplementedError()

    def get_owners(self) -> Union[list[str], Iterable[str]]:
        """
        Returns:
            List of usernames of project owners.
        """
        raise NotImplementedError()

    def who_can_close_issue(self) -> set[str]:
        """
        Returns:
            Names of all users who have permission to modify an issue.
        """
        raise NotImplementedError()

    def who_can_merge_pr(self) -> set[str]:
        """
        Returns:
            Names of all users who have permission to modify pull request.
        """
        raise NotImplementedError()

    def which_groups_can_merge_pr(self) -> set[str]:
        """
        Returns:
            Names of all groups that have permission to modify pull request.
        """
        raise NotImplementedError()

    def can_merge_pr(self, username: str) -> bool:
        """
        Args:
            username: Username.

        Returns:
            `True` if user merge pull request, `False` otherwise.
        """
        raise NotImplementedError()

    def get_users_with_given_access(self, access_levels: list[AccessLevel]) -> set[str]:
        """
        Args:
            access_levels: list of access levels

        Returns:
            set of users with given access levels
        """
        raise NotImplementedError()

    def add_user(self, user: str, access_level: AccessLevel) -> None:
        """
        Add user to project.

        Args:
            user: Username of the user.
            access_level: Permissions for the user.
        """
        raise NotImplementedError()

    def remove_user(self, user: str) -> None:
        """
        Remove user from project.

        Args:
            user: Username of the user.
        """
        raise NotImplementedError()

    def request_access(self) -> None:
        """
        Request an access to the project (cannot specify access level to be granted;
        needs to be approved and specified by the user with maintainer/admin rights).
        """
        raise NotImplementedError()

    def add_group(self, group: str, access_level: AccessLevel) -> None:
        """
        Add group to project.

        Args:
            group: Name of the group.
            access_level: Permissions for the group.
        """
        raise NotImplementedError()

    def remove_group(self, group: str) -> None:
        """
        Remove group from project.

        Args:
            group: Name of the group.
        """
        raise NotImplementedError()

    def get_issue_list(
        self,
        status: IssueStatus = IssueStatus.open,
        author: Optional[str] = None,
        assignee: Optional[str] = None,
        labels: Optional[list[str]] = None,
    ) -> Union[list["_abstract.Issue"], Iterable["_abstract.Issue"]]:
        """
        List of issues.

        Args:
            status: Status of the issues that are to be
                included in the list.

                Defaults to `IssueStatus.open`.
            author: Username of the author of the issues.

                Defaults to no filtering by author.
            assignee: Username of the assignee on the issues.

                Defaults to no filtering by assignees.
            labels: Filter issues that have set specific labels.

                Defaults to no filtering by labels.

        Returns:
            List of objects that represent requested issues.
        """
        raise NotImplementedError()

    def get_issue(self, issue_id: int) -> "_abstract.Issue":
        """
        Get issue.

        Args:
            issue_id: ID of the issue.

        Returns:
            Object that represents requested issue.
        """
        raise NotImplementedError()

    def get_issue_info(self, issue_id: int) -> "_abstract.Issue":
        """
        Get issue info.

        Args:
            issue_id: ID of the issue.

        Returns:
            Object that represents requested issue.
        """
        raise NotImplementedError()

    def create_issue(
        self,
        title: str,
        body: str,
        private: Optional[bool] = None,
        labels: Optional[list[str]] = None,
        assignees: Optional[list[str]] = None,
    ) -> "_abstract.Issue":
        """
        Open new issue.

        Args:
            title: Title of the issue.
            body: Description of the issue.
            private: Is the new issue supposed to be confidential?

                **Supported only by GitLab and Pagure.**

                Defaults to unset.
            labels: List of labels that are to be added to
                the issue.

                Defaults to no labels.
            assignees: List of usernames of the assignees.

                Defaults to no assignees.

        Returns:
            Object that represents newly created issue.

        Raises:
            IssueTrackerDisabled, if issue tracker is disabled.
        """
        raise NotImplementedError()

    def get_pr_list(
        self,
        status: PRStatus = PRStatus.open,
    ) -> Union[list["_abstract.PullRequest"], Iterable["_abstract.PullRequest"]]:
        """
        List of pull requests.

        Args:
            status: Status of the pull requests that are to be included in the list.

                Defaults to `PRStatus.open`.

        Returns:
            List of objects that represent pull requests with requested status.
        """
        raise NotImplementedError()

    def get_pr(self, pr_id: int) -> "_abstract.PullRequest":
        """
        Get pull request.

        Args:
            pr_id: ID of the pull request.

        Returns:
            Object that represents requested pull request.
        """
        raise NotImplementedError()

    def get_pr_files_diff(
        self,
        pr_id: int,
        retries: int = 0,
        wait_seconds: int = 3,
    ) -> dict:
        """
        Get files diff of a pull request.

        Args:
            pr_id: ID of the pull request.

        Returns:
            Dictionary representing files diff.
        """
        raise NotImplementedError()

    def get_tags(self) -> Union[list["GitTag"], Iterable["GitTag"]]:
        """
        Returns:
            List of objects that represent tags.
        """
        raise NotImplementedError()

    def get_sha_from_tag(self, tag_name: str) -> str:
        """
        Args:
            tag_name: Name of the tag.

        Returns:
            Commit hash of the commit from the requested tag.
        """
        raise NotImplementedError()

    def get_release(
        self,
        identifier: Optional[int] = None,
        name: Optional[str] = None,
        tag_name: Optional[str] = None,
    ) -> "_abstract.Release":
        """
        Get a single release.

        Args:
            identifier: Identifier of the release.

                Defaults to `None`, which means not being used.
            name: Name of the release.

                Defaults to `None`, which means not being used.
            tag_name: Tag that the release is tied to.

                Defaults to `None`, which means not being used.

        Returns:
            Object that represents release that satisfies requested condition.
        """
        raise NotImplementedError()

    def get_latest_release(self) -> Optional["_abstract.Release"]:
        """
        Returns:
            Object that represents the latest release.
        """
        raise NotImplementedError()

    def get_releases(
        self,
    ) -> Union[list["_abstract.Release"], Iterable["_abstract.Release"]]:
        """
        Returns:
            List of the objects that represent releases.
        """
        raise NotImplementedError()

    def create_release(
        self,
        tag: str,
        name: str,
        message: str,
        ref: Optional[str] = None,
    ) -> "_abstract.Release":
        """
        Create new release.

        Args:
            tag: Tag which is the release based off.
            name: Name of the release.
            message: Message or description of the release.
            ref: Git reference, mainly commit hash for the release. If provided
                git tag is created prior to creating a release.

                Defaults to `None`.

        Returns:
            Object that represents newly created release.
        """
        raise NotImplementedError()

    def create_pr(
        self,
        title: str,
        body: str,
        target_branch: str,
        source_branch: str,
        fork_username: Optional[str] = None,
    ) -> "_abstract.PullRequest":
        """
        Create new pull request.

        Args:
            title: Title of the pull request.
            body: Description of the pull request.
            target_branch: Name of the branch where the changes are merged.
            source_branch: Name of the branch from which the changes are pulled.
            fork_username: The username of forked repository.

                Defaults to `None`.

        Returns:
            Object that represents newly created pull request.
        """
        raise NotImplementedError()

    def commit_comment(
        self,
        commit: str,
        body: str,
        filename: Optional[str] = None,
        row: Optional[int] = None,
    ) -> "_abstract.CommitComment":
        """
        Add new comment to a commit.

        Args:
            commit: Hash of the commit.
            body: Body of the comment.
            filename: Name of the file that is related to the comment.

                Defaults to `None`, which means no relation to file.
            row: Number of the row that the comment is related to.

                Defaults to `None`, which means no relation to the row.

        Returns:
            Object that represents newly created commit comment.
        """
        raise NotImplementedError()

    def get_commit_comments(
        self,
        commit: str,
    ) -> Union[list["_abstract.CommitComment"], Iterable["_abstract.CommitComment"]]:
        """
        Get comments for a commit.

        Args:
            commit: The hash of the commit.

        Returns:
            List of all comments for the commit.
        """
        raise NotImplementedError()

    def get_commit_comment(
        self,
        commit_sha: str,
        comment_id: int,
    ) -> "_abstract.CommitComment":
        """
        Get commit comment.

        Args:
            commit_sha: SHA of the commit
            comment_id: ID of the commit comment

        Returns:
            Object representing the commit comment.
        """
        raise NotImplementedError()

    def set_commit_status(
        self,
        commit: str,
        state: Union[CommitStatus, str],
        target_url: str,
        description: str,
        context: str,
        trim: bool = False,
    ) -> "_abstract.CommitFlag":
        """
        Create a status on a commit.

        Args:
            commit: The hash of the commit.
            state: The state of the status.
            target_url: The target URL to associate with this status.
            description: A short description of the status.
            context: A label to differentiate this status from the status of other systems.
            trim: Whether to trim the description to 140 characters.

                Defaults to `False`.

        Returns:
            Object that represents created commit status.
        """
        raise NotImplementedError()

    def get_commit_statuses(
        self,
        commit: str,
    ) -> Union[list["_abstract.CommitFlag"], Iterable["_abstract.CommitFlag"]]:
        """
        Get statuses of the commit.

        Args:
            commit: Hash of the commit.

        Returns:
            List of all commit statuses on the commit.
        """
        raise NotImplementedError()

    def get_git_urls(self) -> dict[str, str]:
        """
        Get git URLs for the project.

        Returns:
            Dictionary with at least SSH and HTTP URLs for the current project.
        """
        raise NotImplementedError()

    def fork_create(self, namespace: Optional[str] = None) -> "GitProject":
        """
        Fork this project using the authenticated user.

        Args:
            namespace: Namespace where the project should be forked.

                Defaults to `None`, which means forking to the namespace of
                currently authenticated user.

        Returns:
            Fork of the current project.

        Raises:
            In case the fork already exists.
        """
        raise NotImplementedError()

    def change_token(self, new_token: str) -> None:
        """
        Change an API token. Only for the current instance.

        Args:
            new_token: New token to be set.
        """
        raise NotImplementedError

    def get_file_content(self, path: str, ref: Optional[str] = None) -> str:
        """
        Get a content of the file in the repo.

        Args:
            path: Path to the file.
            ref: Branch or commit.

                Defaults to repo's default branch.

        Returns:
            Contents of the file as string.

        Raises:
            FileNotFoundError: if there is no such file.
        """
        raise NotImplementedError

    def get_files(
        self,
        ref: Optional[str] = None,
        filter_regex: Optional[str] = None,
        recursive: bool = False,
    ) -> Union[list[str], Iterable[str]]:
        """
        Get a list of file paths of the repo.

        Args:
            ref: Branch or commit.

                Defaults to repo's default branch.
            filter_regex: Filter the paths with `re.search`.

                Defaults to `None`, which means no filtering.
            recursive: Whether to return only top directory files
                or all files recursively.

                Defaults to `False`, which means only top-level directory.

        Returns:
            List of paths of the files in the repo.
        """
        raise NotImplementedError

    def get_forks(self) -> Union[Sequence["GitProject"], Iterable["GitProject"]]:
        """
        Returns:
            All forks of the project.
        """
        raise NotImplementedError()

    def get_web_url(self) -> str:
        """
        Returns:
            Web URL of the project.
        """
        raise NotImplementedError()

    def get_sha_from_branch(self, branch: str) -> Optional[str]:
        """
        Returns:
            Commit SHA of head of the branch. `None` if no branch was found.
        """
        raise NotImplementedError()

    def get_contributors(self) -> set[str]:
        """
        Returns:
            Set of all contributors to the given project.
        """
        raise NotImplementedError()

    def users_with_write_access(self) -> set[str]:
        """
        Returns:
            List of users who have write access to the project
        """
        raise NotImplementedError("Use subclass instead.")

    def has_write_access(self, user: str) -> bool:
        """
        Decides whether a given user has write access to the project.

        Args:
            user: The user we are going to check to see if he/she has access
        """
        return user in self.users_with_write_access()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import datetime
import logging
from collections.abc import Iterable
from functools import cached_property, partial
from typing import Optional, Union

import requests
from pyforgejo import NotFoundError
from pyforgejo.types import PullRequest as PyforgejoPullRequest

from ogr.abstract import (
    CommitFlag,
    MergeCommitStatus,
    PRComment,
    PRLabel,
    PRStatus,
    PullRequest,
)
from ogr.exceptions import ForgejoAPIException, OgrNetworkError
from ogr.services import forgejo
from ogr.services.base import BasePullRequest
from ogr.services.forgejo.label import ForgejoPRLabel
from ogr.services.forgejo.utils import paginate

logger = logging.getLogger(__name__)


class ForgejoPullRequest(BasePullRequest):
    _target_project: "forgejo.ForgejoProject" = None
    _source_project: "forgejo.ForgejoProject" = None
    _labels: list[PRLabel] = None

    def __init__(
        self,
        raw_pr: PyforgejoPullRequest,
        project: "forgejo.ForgejoProject",
    ):
        super().__init__(raw_pr, project)

    def __str__(self) -> str:
        return "Forgejo" + super().__str__()

    @property
    def title(self) -> str:
        return self._raw_pr.title

    @title.setter
    def title(self, new_title: str) -> None:
        self.update_info(title=new_title)

    @property
    def id(self) -> int:
        return self._raw_pr.number

    @property
    def status(self) -> PRStatus:
        return PRStatus.merged if self._raw_pr.merged else PRStatus[self._raw_pr.state]

    @property
    def url(self) -> str:
        return self._raw_pr.url

    @property
    def description(self) -> str:
        return self._raw_pr.body

    @description.setter
    def description(self, new_description: str) -> None:
        self.update_info(description=new_description)

    @property
    def author(self) -> str:
        return self._raw_pr.user.login

    @property
    def source_branch(self) -> str:
        return self._raw_pr.head.ref

    @property
    def target_branch(self) -> str:
        return self._raw_pr.base.ref

    @property
    def created(self) -> datetime.datetime:
        return self._raw_pr.created_at

    @property
    def labels(self) -> list[PRLabel]:
        if not self._labels:
            self._labels = (
                [ForgejoPRLabel(raw_label, self) for raw_label in self._raw_pr.labels]
                if self._raw_pr.labels
                else []
            )
        return self._labels

    @property
    def diff_url(self) -> str:
        return self._raw_pr.diff_url

    @property
    def patch(self) -> bytes:
        patch_url = self._raw_pr.patch_url
        response = requests.get(patch_url)

        if not response.ok:
            raise OgrNetworkError(
                f"Couldn't get patch from {patch_url}.patch because {response.reason}.",
            )

        return response.content

    @property
    def head_commit(self) -> str:
        return self._raw_pr.head.sha

    @property
    def merge_commit_sha(self) -> Optional[str]:
        # this is None for non-merged PRs
        return self._raw_pr.merge_commit_sha

    @property
    def merge_commit_status(self) -> MergeCommitStatus:
        return (
            MergeCommitStatus.can_be_merged
            if self._raw_pr.mergeable
            else MergeCommitStatus.cannot_be_merged
        )

    @cached_property
    def source_project(self) -> "forgejo.ForgejoProject":
        pyforgejo_repo = self._raw_pr.head.repo
        return self._target_project.service.get_project(
            repo=pyforgejo_repo.name,
            namespace=pyforgejo_repo.owner.login,
            forgejo_repo=pyforgejo_repo,
        )

    @property
    def commits_url(self) -> str:
        return f"{self.url}/commits"

    @property
    def closed_by(self) -> Optional[str]:
        return self._raw_pr.merged_by.login if self._raw_pr.merged_by else None

    @staticmethod
    def create(
        project: "forgejo.ForgejoProject",
        title: str,
        body: str,
        target_branch: str,
        source_branch: str,
        fork_username: Optional[str] = None,
    ) -> "PullRequest":
        target_project = project

        if project.is_fork and fork_username is None:
            # handles fork -> upstream (called on fork)
            source_branch = f"{project.namespace}:{source_branch}"
            target_project = project.parent  # type: ignore
        elif fork_username:
            if fork_username != project.namespace and project.parent is not None:
                # handles fork -> other_fork
                #   (username of other_fork owner specified by fork_username)
                forks = list(
                    filter(
                        lambda fork: fork.namespace == fork_username,
                        project.parent.get_forks(),
                    ),
                )
                if not forks:
                    raise ForgejoAPIException("Requested fork doesn't exist")
                target_project = forks[0]  # type: ignore
                source_branch = f"{project.namespace}:{source_branch}"
            else:
                # handles fork -> upstream
                #   (username of fork owner specified by fork_username)
                source_branch = f"{fork_username}:{source_branch}"

        logger.debug(f"Creating PR {target_branch}<-{source_branch}")

        pr = target_project.api.repo_create_pull_request(
            owner=target_project.namespace,
            repo=target_project.repo,
            base=target_branch,
            body=body,
            head=source_branch,
            title=title,
        )
        logger.info(f"PR {pr.id} created.")

        return ForgejoPullRequest(pr, target_project)

    @staticmethod
    def get(project: "forgejo.ForgejoProject", pr_id: int) -> "PullRequest":
        try:
            raw_pr = project.api.repo_get_pull_request(
                owner=project.namespace,
                repo=project.repo,
                index=pr_id,
            )
        except NotFoundError as ex:
            raise ForgejoAPIException(f"No pull request with id {pr_id} found.") from ex
        return ForgejoPullRequest(raw_pr, project)

    @staticmethod
    def get_list(
        project: "forgejo.ForgejoProject",
        status: PRStatus = PRStatus.open,
    ) -> Iterable["PullRequest"]:
        prs = paginate(
            partial(
                project.api.repo_list_pull_requests,
                owner=project.namespace,
                repo=project.repo,
                # Forgejo has just open/closed/all
                state=status.name if status != PRStatus.merged else "closed",
            ),
        )
        return (ForgejoPullRequest(pr, project) for pr in prs)

    def update_info(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "PullRequest":
        try:
            data = {"title": title if title else self.title}

            if description is not None:
                data["body"] = description

            updated_pr = self._target_project.api.repo_edit_pull_request(
                owner=self.target_project.namespace,
                repo=self.target_project.repo,
                index=self.id,
                **data,
            )

            self._raw_pr = updated_pr
            return self
        except Exception as ex:
            raise ForgejoAPIException(
                f"There was an error while updating Forgejo PR: {ex}",
            ) from ex

    def close(self) -> "PullRequest":
        self._raw_pr = self._target_project.api.repo_edit_pull_request(
            owner=self.target_project.namespace,
            repo=self.target_project.repo,
            index=self.id,
            state="closed",
        )
        return self

    def merge(self) -> "PullRequest":
        self._target_project.api.repo_merge_pull_request(
            owner=self.target_project.namespace,
            repo=self.target_project.repo,
            index=self.id,
            # options: merge, rebase, rebase-merge, squash, fast-forward-only, manually-merged
            do="merge",
        )
        return self.get(self._target_project, self.id)

    def add_label(self, *labels: str) -> None:
        issue_client = self._target_project.service.api.issue
        new_labels = issue_client.add_label(
            owner=self.target_project.namespace,
            repo=self.target_project.repo,
            index=self.id,
            labels=list(labels),
        )
        self._labels = [ForgejoPRLabel(raw_label, self) for raw_label in new_labels]

    def get_all_commits(self) -> Iterable[str]:
        return (
            commit.sha
            for commit in paginate(
                partial(
                    self._target_project.api.repo_get_pull_request_commits,
                    owner=self.target_project.namespace,
                    repo=self.target_project.repo,
                    index=self.id,
                ),
            )
        )

    def get_comments(
        self,
        filter_regex: Optional[str] = None,
        reverse: bool = False,
        author: Optional[str] = None,
    ) -> Union[list["PRComment"], Iterable["PRComment"]]:
        """
        Get list of pull request comments.

        Args:
            filter_regex: Filter the comments' content with `re.search`.

                Defaults to `None`, which means no filtering.
            reverse: Whether the comments are to be returned in
                reversed order.

                Defaults to `False`.
            author: Filter the comments by author.

                Defaults to `None`, which means no filtering.

        Returns:
            List of pull request comments.
        """
        raise NotImplementedError()

    def comment(
        self,
        body: str,
        commit: Optional[str] = None,
        filename: Optional[str] = None,
        row: Optional[int] = None,
    ) -> "PRComment":
        """
        Add new comment to the pull request.

        Args:
            body: Body of the comment.
            commit: Commit hash to which comment is related.

                Defaults to generic comment.
            filename: Path to the file to which comment is related.

                Defaults to no relation to the file.
            row: Line number to which the comment is related.

                Defaults to no relation to the line.

        Returns:
            Newly created comment.
        """
        raise NotImplementedError()

    def get_comment(self, comment_id: int) -> PRComment:
        """
        Returns a PR comment.

        Args:
            comment_id: id of comment

        Returns:
            Object representing a PR comment.
        """
        raise NotImplementedError()

    def get_statuses(self) -> Union[list[CommitFlag], Iterable[CommitFlag]]:
        """
        Returns statuses for latest commit on pull request.

        Returns:
            List of commit statuses of the latest commit.
        """
        raise NotImplementedError()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Optional

from pyforgejo.types import Issue as PyforgejoIssue

from ogr.abstract import Issue, IssueStatus
from ogr.services import forgejo
from ogr.services.base import BaseIssue


class ForgejoIssue(BaseIssue):
    def __init__(self, raw_issue: PyforgejoIssue, project: "forgejo.ForgejoProject"):
        super().__init__(raw_issue, project)

    @staticmethod
    def create(
        project: "forgejo.ForgejoProject",
        title: str,
        body: str,
        private: Optional[bool] = None,
        labels: Optional[list[str]] = None,
        assignees: Optional[list] = None,
    ) -> "Issue":
        raise NotImplementedError("TBD")

    @staticmethod
    def get(project: "forgejo.ForgejoProject", issue_id: int) -> "Issue":
        raise NotImplementedError("TBD")

    @staticmethod
    def get_list(
        project: "forgejo.ForgejoProject",
        status: IssueStatus = IssueStatus.open,
        author: Optional[str] = None,
        assignee: Optional[str] = None,
        labels: Optional[list[str]] = None,
    ) -> list["Issue"]:
        raise NotImplementedError("TBD")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from ogr.abstract.abstract_class import OgrAbstractClass


class GitTag(OgrAbstractClass):
    """
    Class representing a git tag.

    Attributes:
        name (str): Name of the tag.
        commit_sha (str): Commit hash of the tag.
    """

    def __init__(self, name: str, commit_sha: str) -> None:
        self.name = name
        self.commit_sha = commit_sha

    def __str__(self) -> str:
        return f"GitTag(name={self.name}, commit_sha={self.commit_sha})"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import codecs
import contextlib
import logging
from collections.abc import Iterable
from functools import cached_property, partial
from typing import ClassVar, Optional, Union

from pyforgejo import NotFoundError, Repository, types
from pyforgejo.repository.client import RepositoryClient

from ogr.abstract import (
    AccessLevel,
    CommitComment,
    CommitFlag,
    CommitStatus,
    GitProject,
    GitTag,
    Issue,
    IssueStatus,
    PRStatus,
    PullRequest,
    Release,
)
from ogr.exceptions import OperationNotSupported
from ogr.services import forgejo
from ogr.services.base import BaseGitProject
from ogr.utils import filter_paths, indirect

from .flag import ForgejoCommitFlag
from .issue import ForgejoIssue
from .pull_request import ForgejoPullRequest
from .release import ForgejoRelease
from .utils import paginate

logger = logging.getLogger(__name__)


class ForgejoProject(BaseGitProject):
    service: "forgejo.ForgejoService"
    access_dict: ClassVar[dict] = {
        AccessLevel.pull: "read",
        AccessLevel.triage: "read",
        AccessLevel.push: "write",
        AccessLevel.admin: "admin",
        AccessLevel.maintain: "owner",
        None: "",
    }

    def __init__(
        self,
        repo: str,
        service: "forgejo.ForgejoService",
        namespace: str,
        forgejo_repo: Optional[Repository] = None,
        **kwargs,
    ):
        super().__init__(repo, service, namespace)
        self._forgejo_repo = forgejo_repo

    @property
    def api(self) -> RepositoryClient:
        """Returns a `RepositoryClient` from pyforgejo. Helper to save some
        typing.
        """
        return self.service.api.repository

    def partial_api(self, method, /, *args, **kwargs):
        """Returns a partial API call for `ForgejoProject`.

        Injects `owner` and `repo` for the calls to `/repository/` endpoints.

        Args:
            method: Specific method on the Pyforgejo API that is to be wrapped.
            *args: Positional arguments that get injected into every call.
            **kwargs: Keyword-arguments that get injected into every call.

        Returns:
            Callable with pre-injected parameters.

        """
        return partial(
            method,
            *args,
            **kwargs,
            owner=self.namespace,
            repo=self.repo,
        )

    @cached_property
    def forgejo_repo(self) -> types.Repository:
        return self.api.repo_get(
            owner=self.namespace,
            repo=self.repo,
        )

    def __str__(self) -> str:
        return (
            f'ForgejoProject(namespace="{self.namespace}", repo="{self.repo}", '
            f"service={self.service})"
        )

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, ForgejoProject)
            and self.repo == o.repo
            and self.namespace == o.namespace
            and self.service == o.service
        )

    @property
    def description(self) -> str:
        return self.forgejo_repo.description or ""

    @description.setter
    def description(self, new_description: str) -> None:
        self.partial_api(self.api.repo_edit)(description=new_description)

    def delete(self) -> None:
        self.partial_api(self.api.repo_delete)()

    def exists(self) -> bool:
        try:
            _ = self.forgejo_repo
            return True
        except NotFoundError:
            return False

    def is_private(self) -> bool:
        return self.forgejo_repo.private

    def is_forked(self) -> bool:
        return (
            self.forgejo_repo.fork
            and self.forgejo_repo.owner.login == self.service.user.get_username()
        )

    @property
    def is_fork(self) -> bool:
        return self.forgejo_repo.fork

    @property
    def full_repo_name(self) -> str:
        return self.forgejo_repo.full_name

    @property
    def parent(self) -> Optional["GitProject"]:
        if not self.forgejo_repo.parent:
            return None

        return ForgejoProject(
            service=self.service,
            repo=self.forgejo_repo.parent.name,
            namespace=self.forgejo_repo.parent.owner.username,
        )

    @property
    def has_issues(self) -> bool:
        return self.forgejo_repo.has_issues

    def get_branches(self) -> Iterable[str]:
        return (
            branch.name
            for branch in paginate(
                self.partial_api(self.api.repo_list_branches),
            )
        )

    @property
    def default_branch(self) -> str:
        return self.forgejo_repo.default_branch

    def get_commits(self, ref: Optional[str] = None) -> Iterable[str]:
        return (
            commit.sha
            for commit in paginate(
                self.partial_api(
                    self.api.repo_get_all_commits,
                    sha=ref,
                ),
            )
        )

    def get_description(self) -> str:
        return self.description

    def _construct_fork_project(self) -> Optional["ForgejoProject"]:
        login = self.service.user.get_username()
        try:
            project = ForgejoProject(
                repo=self.repo,
                service=self.service,
                namespace=login,
            )
            _ = project.forgejo_repo
            return project
        except NotFoundError:
            return None

    def get_fork(self, create: bool = True) -> Optional["GitProject"]:
        # The cheapest check that assumes fork has the same repository name as
        # the upstream
        if fork := self._construct_fork_project():
            return fork

        # If not successful, the fork could still exist, but has a custom name
        username = self.service.user.get_username()
        for fork in self.get_forks():
            if fork.forgejo_repo.owner.login == username:
                return fork

        # We have not found any fork owned by the auth'd user
        if create:
            return self.fork_create()

        logger.info(
            f"Fork of {self.forgejo_repo.full_name}"
            " does not exist and we were asked not to create it.",
        )
        return None

    def get_owners(self) -> list[str]:
        return [self.forgejo_repo.owner.username]

    def _get_owner_or_org_collaborators(self) -> set[str]:
        namespace = self.get_owners()[0]
        try:
            teams = self.api.repo_list_teams(
                owner=self.namespace,
                repo=self.repo,
            )
        except Exception as ex:
            # no teams, repo owned by regular user
            if "not owned by an organization" in str(ex):
                return {namespace}
            raise

        # repo owned by org, each org can have multiple teams with
        # different levels of access
        collaborators: set[str] = set()
        for team in teams:
            members = self.service.api.organization.org_list_team_members(team.id)
            collaborators.update(user.username for user in members)

        return collaborators

    def _get_collaborators(self) -> list[str]:
        return [
            c.username
            for c in self.api.repo_list_collaborators(
                owner=self.namespace,
                repo=self.repo,
            )
        ] + list(self._get_owner_or_org_collaborators())

    def _get_collaborators_with_access(self) -> dict[str, str]:
        return {
            c: self.api.repo_get_repo_permissions(
                owner=self.namespace,
                repo=self.repo,
                collaborator=c,
            ).permission
            for c in self._get_collaborators()
        }

    def get_contributors(self) -> set[str]:
        return set(self._get_collaborators())

    def users_with_write_access(self) -> set[str]:
        return {
            collaborator
            for collaborator, access in self._get_collaborators_with_access().items()
            if access in ("owner", "admin", "write")
        }

    def who_can_close_issue(self) -> set[str]:
        return self.users_with_write_access()

    def who_can_merge_pr(self) -> set[str]:
        return self.users_with_write_access()

    def can_merge_pr(self, username: str) -> bool:
        return self.api.repo_get_repo_permissions(
            owner=self.namespace,
            repo=self.repo,
            collaborator=username,
        ).permission in ("owner", "admin", "write")

    def get_users_with_given_access(self, access_levels: list[AccessLevel]) -> set[str]:
        access_levels_forgejo = [
            self.access_dict[access_level] for access_level in access_levels
        ]

        return {
            user
            for user, permission in self._get_collaborators_with_access().items()
            if permission in access_levels_forgejo
        }

    def add_user(self, user: str, access_level: AccessLevel) -> None:
        if access_level == AccessLevel.maintain:
            raise OperationNotSupported("Not possible to add a user as `owner`.")

        self.api.repo_add_collaborator(
            owner=self.namespace,
            repo=self.repo,
            collaborator=user,
            permission=self.access_dict[access_level],
        )

    def remove_user(self, user: str) -> None:
        self.api.repo_delete_collaborator(
            owner=self.namespace,
            repo=self.repo,
            collaborator=user,
        )

    def request_access(self) -> None:
        raise OperationNotSupported("Not possible on Forgejo")

    @indirect(ForgejoIssue.get_list)
    def get_issue_list(
        self,
        status: IssueStatus = IssueStatus.open,
        author: Optional[str] = None,
        assignee: Optional[str] = None,
        labels: Optional[list[str]] = None,
    ) -> list["Issue"]:
        pass

    @indirect(ForgejoIssue.get)
    def get_issue(self, issue_id: int) -> "Issue":
        pass

    @indirect(ForgejoIssue.create)
    def create_issue(
        self,
        title: str,
        body: str,
        private: Optional[bool] = None,
        labels: Optional[list[str]] = None,
        assignees: Optional[list[str]] = None,
    ) -> Issue:
        pass

    @indirect(ForgejoPullRequest.get_list)
    def get_pr_list(self, status: PRStatus = PRStatus.open) -> Iterable["PullRequest"]:
        pass

    @indirect(ForgejoPullRequest.get)
    def get_pr(self, pr_id: int) -> "PullRequest":
        pass

    def get_pr_files_diff(
        self,
        pr_id: int,
        retries: int = 0,
        wait_seconds: int = 3,
    ) -> dict:
        """
        Get files diff of a pull request.

        Args:
            pr_id: ID of the pull request.

        Returns:
            Dictionary representing files diff.
        """
        # [NOTE] Implemented only for Pagure, for details see
        # https://github.com/packit/ogr/issues/895
        raise NotImplementedError()

    def get_tags(self) -> Iterable["GitTag"]:
        return (
            GitTag(
                name=tag.name,
                commit_sha=tag.commit.sha,
            )
            for tag in paginate(self.partial_api(self.api.repo_list_tags))
        )

    def get_sha_from_tag(self, tag_name: str) -> str:
        return self.partial_api(
            self.api.repo_get_tag,
            tag=tag_name,
        )().commit.sha

    @indirect(ForgejoRelease.get)
    def get_release(
        self,
        identifier: Optional[int] = None,
        name: Optional[str] = None,
        tag_name: Optional[str] = None,
    ) -> Release:
        pass

    @indirect(ForgejoRelease.get_latest)
    def get_latest_release(self) -> Optional[Release]:
        pass

    @indirect(ForgejoRelease.get_list)
    def get_releases(self) -> list[Release]:
        pass

    @indirect(ForgejoRelease.create)
    def create_release(
        self,
        tag: str,
        name: str,
        message: str,
        ref: Optional[str] = None,
    ) -> Release:
        pass

    @indirect(ForgejoPullRequest.create)
    def create_pr(
        self,
        title: str,
        body: str,
        target_branch: str,
        source_branch: str,
        fork_username: Optional[str] = None,
    ) -> "PullRequest":
        pass

    def commit_comment(
        self,
        commit: str,
        body: str,
        filename: Optional[str] = None,
        row: Optional[int] = None,
    ) -> "CommitComment":
        raise OperationNotSupported("Forgejo doesn't support commit comments")

    def get_commit_comments(self, commit: str) -> list[CommitComment]:
        raise OperationNotSupported("Forgejo doesn't support commit comments")

    def get_commit_comment(self, commit_sha: str, comment_id: int) -> CommitComment:
        raise OperationNotSupported("Forgejo doesn't support commit comments")

    @indirect(ForgejoCommitFlag.set)
    def set_commit_status(
        self,
        commit: str,
        state: Union[CommitStatus, str],
        target_url: str,
        description: str,
        context: str,
        trim: bool = False,
    ) -> "CommitFlag":
        pass

    @indirect(ForgejoCommitFlag.get)
    def get_commit_statuses(self, commit: str) -> Iterable["CommitFlag"]:
        pass

    def get_git_urls(self) -> dict[str, str]:
        return {
            "git": self.forgejo_repo.clone_url,
            "ssh": self.forgejo_repo.ssh_url,
        }

    def fork_create(self, namespace: Optional[str] = None) -> "GitProject":
        if namespace:
            self.api.create_fork(
                owner=self.namespace,
                repo=self.repo,
                organization=namespace,
            )
            return ForgejoProject(
                repo=self.repo,
                service=self.service,
                namespace=namespace,
            )

        self.api.create_fork(
            owner=self.namespace,
            repo=self.repo,
        )
        return ForgejoProject(
            repo=self.repo,
            service=self.service,
            namespace=self.service.user.get_username(),
        )

    def change_token(self, new_token: str) -> None:
        # [NOTE] API doesn't provide any method to change the token, and it's
        # embedded in the httpx client that's wrapped by pyforgejo wrapper to
        # avoid duplication between sync and async calls…
        raise NotImplementedError(
            "Not possible; requires recreation of the httpx client",
        )

    def get_file_content(self, path: str, ref: Optional[str] = None) -> str:
        try:
            remote_file: types.ContentsResponse = self.partial_api(
                self.api.repo_get_contents,
                filepath=path,
                ref=ref,
            )()

            # [NOTE] If you touch this, good luck, have fun…
            # tl;dr ‹ContentsResponse› from the Pyforgejo contains the content
            # of the file that's (I hope always) base64-encoded, but it's stored
            # as a string, so here it's needed to convert the UTF-8 encoded
            # string back to bytes (duh, cause base64 is used for encoding raw
            # data), then decode the base64 bytes to just bytes and then decode
            # those to a UTF-8 string… EWWW…
            return codecs.decode(
                bytes(remote_file.content, "utf-8"),
                encoding=remote_file.encoding,
            ).decode("utf-8")

        except NotFoundError as ex:
            raise FileNotFoundError() from ex

    def __get_files(
        self,
        path: str,
        ref: str,
        recursive: bool,
    ) -> Iterable[str]:
        contents: types.ContentsResponse | list[types.ContentsResponse]

        subdirectories = ["."]

        with contextlib.suppress(IndexError):
            while path := subdirectories.pop():
                contents = self.partial_api(
                    self.api.repo_get_contents,
                    filepath=path,
                    ref=ref,
                )()

                if isinstance(contents, types.ContentsResponse):
                    # singular file, return path and skip any further processing
                    yield contents.path
                    continue

                for file in contents:
                    if file.type == "dir":
                        subdirectories.append(file.path)
                        continue

                    yield file.path

    def get_files(
        self,
        ref: Optional[str] = None,
        filter_regex: Optional[str] = None,
        recursive: bool = False,
    ) -> Iterable[str]:
        logger.warning(
            "‹ForgejoProject.get_files()› method can fail because of incorrect"
            " OpenAPI spec",
        )

        ref = ref or self.default_branch
        paths = self.__get_files(".", ref=ref, recursive=recursive)

        if filter_regex:
            return filter_paths(paths, filter_regex)

        return paths

    def get_forks(self) -> Iterable["ForgejoProject"]:
        return (
            ForgejoProject(
                namespace=fork.owner.login,
                repo=fork.name,
                service=self.service,
            )
            for fork in paginate(
                self.partial_api(self.api.list_forks),
            )
        )

    def get_web_url(self) -> str:
        return self.forgejo_repo.html_url

    def get_sha_from_branch(self, branch: str) -> Optional[str]:
        try:
            branch_info = self.partial_api(
                self.api.repo_get_branch,
                branch=branch,
            )()
            return branch_info.commit.id
        except NotFoundError:
            return None



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from requre.helpers import record_httpx


@record_httpx()
def test_get_email(service):
    assert service.user.get_email() == "public-repos-bot@packit.dev"


@record_httpx()
def test_get_projects(service):
    """When recording this test,
    run it after test_service.py which will create the repos.
    """
    projects = service.user.get_projects()
    assert len(projects) == 3
    namespaces = [project.namespace for project in projects]
    repos = [project.repo for project in projects]
    assert "packit" in namespaces
    assert "packit-validator" in namespaces
    assert "test" in repos
    assert "test_1" in repos


@record_httpx()
def test_get_forks(service):
    forks = service.user.get_forks()
    assert len(forks) == 0


@record_httpx()
def test_username(service):
    # check just length, because it is based who regenerated data files
    assert len(service.user.get_username()) > 3


@record_httpx()
def test_email(service):
    email = service.user.get_email()
    assert email
    assert len(email) > 3
    assert "@" in email
    assert "." in email



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from requre.helpers import record_httpx

from ogr.abstract import MergeCommitStatus, PRStatus
from ogr.exceptions import ForgejoAPIException


@record_httpx()
def test_pr_list(project):
    pr_list = list(project.get_pr_list(status=PRStatus.all))
    assert len({pr.id for pr in pr_list}) > 20


@record_httpx()
def test_pr_info(project):
    pr_info = project.get_pr(pr_id=119)
    assert pr_info.id == 119
    assert pr_info.title == "change"
    assert pr_info.description == "description of mergerequest"
    assert pr_info.url.startswith("https://v10.next.forgejo.org")
    assert pr_info.source_branch == "change"
    assert pr_info.target_branch == "master"
    assert pr_info.status == PRStatus.merged


@record_httpx()
def test_pr_not_exists(project):
    with pytest.raises(ForgejoAPIException):
        project.get_pr(2000)


@record_httpx()
def test_commits_url(project):
    pr = project.get_pr(119)
    assert (
        pr.commits_url
        == "https://v10.next.forgejo.org/packit-validator/ogr-tests/pulls/119/commits"
    )


@record_httpx()
def test_pr_patch(project):
    pr = project.get_pr(119)
    patch = pr.patch
    assert isinstance(patch, bytes)
    assert (
        "From: Laura Barcziova <lbarczio@redhat.com>\nDate: Tue, 10 Sep 2019 15:54:50 +0200\n"
        in patch.decode()
    )


@record_httpx()
def test_get_all_pr_commits(project):
    commits = list(project.get_pr(119).get_all_commits())
    assert commits == [
        "d490ec67dd98f69dfdc1732b98bb3189f0e0aace",
        "3c1fb11dd358254cc3f1588f173e54e98c1d4c09",
    ]


@record_httpx()
def test_pr_labels(project):
    # remove the labels before generating
    pr = project.get_pr(209)
    labels = pr.labels
    assert not labels

    pr.add_label("test_lb1", "test_lb2")
    labels = pr.labels

    assert {label.name for label in labels} >= {"test_lb1", "test_lb2"}


@record_httpx()
def test_head_commit(project):
    assert project.get_pr(119).head_commit == "d490ec67dd98f69dfdc1732b98bb3189f0e0aace"
    assert project.get_pr(137).head_commit == "59b1a9bab5b5198c619270646410867788685c16"


@record_httpx()
def test_target_branch_head_commit(project):
    assert (
        project.get_pr(204).target_branch_head_commit
        == "dd9b6c54c0788301c86bdf058a8e91e3594a0a17"
    )


@record_httpx()
def test_setters(project):
    pr = project.get_pr(119)
    old_title, old_description = pr.title, pr.description
    pr.title = "test title"
    assert pr.title == "test title"
    pr.title = old_title
    pr.description = "test description"
    assert pr.description == "test description"
    pr.description = old_description


@record_httpx()
def test_merge_commit_sha(project):
    pr = project.get_pr(189)
    assert pr.merge_commit_status == MergeCommitStatus.cannot_be_merged
    assert pr.merge_commit_sha == "dd9b6c54c0788301c86bdf058a8e91e3594a0a17"


@record_httpx()
def test_pr_create_upstream_upstream(project):
    pr_opened_before = len(list(project.get_pr_list(status=PRStatus.open)))
    pr = project.create_pr(
        title="test: upstream <- upstream",
        body="pull request body",
        target_branch="master",
        source_branch="readme-change-1",
    )
    pr_opened_after = len(list(project.get_pr_list(status=PRStatus.open)))

    assert pr.title == "test: upstream <- upstream"
    assert pr.status == PRStatus.open
    assert not pr.target_project.is_fork
    assert pr_opened_after == pr_opened_before + 1

    pr.close()
    assert pr.status == PRStatus.closed


@record_httpx()
def test_pr_create_upstream_forkusername(project):
    pr_opened_before = len(list(project.get_pr_list(status=PRStatus.open)))
    pr = project.create_pr(
        title="test: upstream <- fork_username:source_branch",
        body="pull request body",
        target_branch="master",
        source_branch="readme-change-1",
        fork_username="lbarcziova",
    )
    pr_opened_after = len(list(project.get_pr_list(status=PRStatus.open)))

    assert pr.title == "test: upstream <- fork_username:source_branch"
    assert pr.status == PRStatus.open
    assert not pr.target_project.is_fork
    assert pr_opened_after == pr_opened_before + 1

    pr.close()
    assert pr.status == PRStatus.closed


@record_httpx()
def test_pr_create_upstream_fork(project):
    fork_project = project.service.get_project(namespace="lbarcziova", repo="ogr-tests")
    pr_opened_before = len(list(project.get_pr_list(status=PRStatus.open)))
    pr = fork_project.create_pr(
        title="test: upstream <- fork",
        body="pull request body",
        target_branch="master",
        source_branch="readme-change-1",
    )
    pr_opened_after = len(list(project.get_pr_list(status=PRStatus.open)))

    assert pr.title == "test: upstream <- fork"
    assert pr.status == PRStatus.open
    assert not pr.target_project.is_fork
    assert pr_opened_after == pr_opened_before + 1

    pr.close()
    assert pr.status == PRStatus.closed


@record_httpx
def test_pr_create_fork_fork(project):
    fork_project = project.service.get_project(namespace="lbarcziova", repo="ogr-tests")

    pr_opened_before = len(list(fork_project.get_pr_list(status=PRStatus.open)))
    pr = fork_project.create_pr(
        title="test: fork(master) <- fork",
        body="pull request body",
        target_branch="master",
        source_branch="readme-change-1",
        fork_username="lbarcziova",
    )
    pr_opened_after = len(list(fork_project.get_pr_list(status=PRStatus.open)))

    assert pr.title == "test: fork(master) <- fork"
    assert pr.status == PRStatus.open
    assert pr.target_project.is_fork
    assert pr_opened_after == pr_opened_before + 1

    pr.close()
    assert pr.status == PRStatus.closed


@record_httpx()
def test_source_project(project):
    pr = project.get_pr(209)
    assert pr.source_project.repo == "ogr-tests"
    assert pr.source_project.namespace == "packit-validator"
    assert not pr.source_project.is_fork


@record_httpx()
def test_source_project_fork(project):
    pr = project.get_pr(211)
    assert pr.source_project.repo == "ogr-tests"
    assert pr.source_project.namespace == "lbarcziova"
    assert pr.source_project.is_fork



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from collections.abc import Iterable
from typing import Any, Optional, Union

from ogr import abstract as _abstract
from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.auth_method import AuthMethod
from ogr.abstract.git_project import GitProject
from ogr.exceptions import OgrException
from ogr.parsing import parse_git_repo

try:
    from functools import cached_property as _cached_property  # type: ignore
except ImportError:
    from functools import lru_cache

    def _cached_property(func):  # type: ignore
        return property(lru_cache()(func))


class GitService(OgrAbstractClass):
    """
    Attributes:
        instance_url (str): URL of the git forge instance.
    """

    instance_url: Optional[str] = None

    def __init__(self, **_: Any) -> None:
        pass

    def __str__(self) -> str:
        return f"GitService(instance_url={self.instance_url})"

    def get_project(self, **kwargs: Any) -> "GitProject":
        """
        Get the requested project.

        Args:
            namespace (str): Namespace of the project.
            user (str): Username of the project's owner.
            repo (str): Repository name.

        Returns:
            Object that represents git project.
        """
        raise NotImplementedError

    def get_project_from_url(self, url: str) -> "GitProject":
        """
        Args:
            url: URL of the git repository.

        Returns:
            Object that represents project from the parsed URL.
        """
        repo_url = parse_git_repo(potential_url=url)
        if not repo_url:
            raise OgrException(f"Failed to find repository for url: {url}")
        return self.get_project(repo=repo_url.repo, namespace=repo_url.namespace)

    @_cached_property
    def hostname(self) -> Optional[str]:
        """Hostname of the service."""
        raise NotImplementedError

    @property
    def user(self) -> "_abstract.GitUser":
        """User authenticated through the service."""
        raise NotImplementedError

    def change_token(self, new_token: str) -> None:
        """
        Change an API token. Only for the current instance and newly created projects.

        Args:
            new_token: New token to be set.
        """
        raise NotImplementedError

    def set_auth_method(self, method: AuthMethod) -> None:
        """
        Override the default auth method.
        Can be used when the service has more auth methods available.

        Args:
            method: the method identifier (a str name)
        """
        raise NotImplementedError()

    def reset_auth_method(self) -> None:
        """
        Set the auth method to the default one.
        """
        raise NotImplementedError()

    def project_create(
        self,
        repo: str,
        namespace: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "GitProject":
        """
        Create new project.

        Args:
            repo: Name of the newly created project.
            namespace: Namespace of the newly created project.

                Defaults to currently authenticated user.
            description: Description of the newly created project.

        Returns:
            Object that represents newly created project.
        """
        raise NotImplementedError()

    def list_projects(
        self,
        namespace: Optional[str] = None,
        user: Optional[str] = None,
        search_pattern: Optional[str] = None,
        language: Optional[str] = None,
    ) -> Union[list["GitProject"], Iterable["GitProject"]]:
        """
        List projects for given criteria.

        Args:
            namespace: Namespace to list projects from.
            user: Login of the owner of the projects.
            search_pattern: Regular expression that repository name should match.
            language: Language to be present in the project, e.g. `"python"` or
                `"html"`.
        """
        raise NotImplementedError

    def get_group(self, group_name: str):
        """
        Get a group by name.
        """
        raise NotImplementedError



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from enum import IntEnum


class AccessLevel(IntEnum):
    """
    Enumeration representing an access level to the repository.

    | Value from enumeration | GitHub   | GitLab                  | Pagure |
    | ---------------------- | -------- | ----------------------- | ------ |
    | `AccessLevel.pull`     | pull     | guest                   | ticket |
    | `AccessLevel.triage`   | triage   | reporter                | ticket |
    | `AccessLevel.push`     | push     | developer               | commit |
    | `AccessLevel.admin`    | admin    | maintainer              | commit |
    | `AccessLevel.maintain` | maintain | owner (only for groups) | admin  |
    """

    pull = 1
    triage = 2
    push = 3
    admin = 4
    maintain = 5



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Any

from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.issue import Issue
from ogr.abstract.pull_request import PullRequest


class Label(OgrAbstractClass):
    """
    Represents labels on PRs and issues.
    """

    def __init__(self, parent: Any) -> None:
        self._parent = parent

    @property
    def name(self) -> str:
        """Name of the label."""
        raise NotImplementedError()


class IssueLabel(Label):
    @property
    def issue(self) -> "Issue":
        """Issue of issue label."""
        return self._parent

    def __str__(self) -> str:
        return "Issue" + super().__str__()


class PRLabel(Label):
    @property
    def pull_request(self) -> "PullRequest":
        """Pull request of pull request label."""
        return self._parent

    def __str__(self) -> str:
        return "PR" + super().__str__()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from collections.abc import Iterable
from typing import Any, Optional, Union

from ogr import abstract as _abstract
from ogr.abstract.abstract_class import OgrAbstractClass
from ogr.abstract.status import IssueStatus


class Issue(OgrAbstractClass):
    """
    Attributes:
        project (GitProject): Project of the issue.
    """

    def __init__(self, raw_issue: Any, project: "_abstract.GitProject") -> None:
        self._raw_issue = raw_issue
        self.project = project

    @property
    def title(self) -> str:
        """Title of the issue."""
        raise NotImplementedError()

    @property
    def private(self) -> bool:
        """`True` if issue is confidential, `False` otherwise."""
        raise NotImplementedError()

    @property
    def id(self) -> int:
        """ID of the issue."""
        raise NotImplementedError()

    @property
    def status(self) -> IssueStatus:
        """Status of the issue."""
        raise NotImplementedError()

    @property
    def url(self) -> str:
        """Web URL of the issue."""
        raise NotImplementedError()

    @property
    def description(self) -> str:
        """Description of the issue."""
        raise NotImplementedError()

    @property
    def author(self) -> str:
        """Username of the author of the issue."""
        raise NotImplementedError()

    @property
    def created(self) -> datetime.datetime:
        """Datetime of the creation of the issue."""
        raise NotImplementedError()

    @property
    def labels(
        self,
    ) -> Union[list["_abstract.IssueLabel"], Iterable["_abstract.IssueLabel"]]:
        """Labels of the issue."""
        raise NotImplementedError()

    def __str__(self) -> str:
        description = (
            f"{self.description[:10]}..." if self.description is not None else "None"
        )
        return (
            f"Issue("
            f"title='{self.title}', "
            f"id={self.id}, "
            f"status='{self.status.name}', "
            f"url='{self.url}', "
            f"description='{description}', "
            f"author='{self.author}', "
            f"created='{self.created}')"
        )

    @staticmethod
    def create(
        project: Any,
        title: str,
        body: str,
        private: Optional[bool] = None,
        labels: Optional[list[str]] = None,
        assignees: Optional[list[str]] = None,
    ) -> "Issue":
        """
        Open new issue.

        Args:
            project (GitProject): Project where the issue is to be opened.
            title: Title of the issue.
            body: Description of the issue.
            private: Is the new issue supposed to be confidential?

                **Supported only by GitLab and Pagure.**

                Defaults to unset.
            labels: List of labels that are to be added to
                the issue.

                Defaults to no labels.
            assignees: List of usernames of the assignees.

                Defaults to no assignees.

        Returns:
            Object that represents newly created issue.
        """
        raise NotImplementedError()

    @staticmethod
    def get(project: Any, id: int) -> "Issue":
        """
        Get issue.

        Args:
            project (GitProject): Project where the issue is to be opened.
            issue_id: ID of the issue.

        Returns:
            Object that represents requested issue.
        """
        raise NotImplementedError()

    @staticmethod
    def get_list(
        project: Any,
        status: IssueStatus = IssueStatus.open,
        author: Optional[str] = None,
        assignee: Optional[str] = None,
        labels: Optional[list[str]] = None,
    ) -> Union[list["Issue"], Iterable["Issue"]]:
        """
        List of issues.

        Args:
            project (GitProject): Project where the issue is to be opened.
            status: Status of the issues that are to be
                included in the list.

                Defaults to `IssueStatus.open`.
            author: Username of the author of the issues.

                Defaults to no filtering by author.
            assignee: Username of the assignee on the issues.

                Defaults to no filtering by assignees.
            labels: Filter issues that have set specific labels.

                Defaults to no filtering by labels.

        Returns:
            List of objects that represent requested issues.
        """
        raise NotImplementedError()

    def _get_all_comments(
        self,
        reverse: bool = False,
    ) -> Union[list["_abstract.IssueComment"], Iterable["_abstract.IssueComment"]]:
        """
        Get list of all issue comments.

        Args:
            reverse: Defines whether the comments should be listed in a reversed
                order.

                Defaults to `False`.

        Returns:
            List of all comments on the issue.
        """
        raise NotImplementedError()

    def get_comments(
        self,
        filter_regex: Optional[str] = None,
        reverse: bool = False,
        author: Optional[str] = None,
    ) -> Union[list["_abstract.IssueComment"], Iterable["_abstract.IssueComment"]]:
        """
        Get list of issue comments.

        Args:
            filter_regex: Filter the comments' content with `re.search`.

                Defaults to `None`, which means no filtering.
            reverse: Whether the comments are to be returned in
                reversed order.

                Defaults to `False`.
            author: Filter the comments by author.

                Defaults to `None`, which means no filtering.

        Returns:
            List of issue comments.
        """
        raise NotImplementedError()

    def can_close(self, username: str) -> bool:
        """
        Check if user have permissions to modify an issue.

        Args:
            username: Login of the user.

        Returns:
            `True` if user can close the issue, `False` otherwise.
        """
        raise NotImplementedError()

    def comment(self, body: str) -> "_abstract.IssueComment":
        """
        Add new comment to the issue.

        Args:
            body: Text contents of the comment.

        Returns:
            Object that represents posted comment.
        """
        raise NotImplementedError()

    def close(self) -> "Issue":
        """
        Close an issue.

        Returns:
            Issue itself.
        """
        raise NotImplementedError()

    def add_label(self, *labels: str) -> None:
        """
        Add labels to the issue.

        Args:
            *labels: Labels to be added.
        """
        raise NotImplementedError()

    def add_assignee(self, *assignees: str) -> None:
        """
        Assign users to an issue.

        Args:
            *assignees: List of logins of the assignees.
        """
        raise NotImplementedError()

    def get_comment(self, comment_id: int) -> "_abstract.IssueComment":
        """
        Returns an issue comment.

        Args:
            comment_id: id of a comment

        Returns:
            Object representing an issue comment.
        """
        raise NotImplementedError()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from requre.helpers import record_httpx


@record_httpx
def test_get_release(project):
    release = project.get_release(tag_name="0.1.0")
    assert release.title == "test"
    assert release.body == "testing release"


@record_httpx
def test_get_releases(project):
    releases = project.get_releases()
    assert releases
    assert len(releases) >= 9


@record_httpx
def test_create_release(project):
    releases_before = project.get_releases()
    latest_release = releases_before[0].tag_name
    count_before = len(releases_before)
    increased_release = ".".join(
        [
            latest_release.rsplit(".", 1)[0],
            str(int(latest_release.rsplit(".", 1)[1]) + 1),
        ],
    )
    release = project.create_release(
        tag=increased_release,
        name="test",
        message="testing release",
    )
    count_after = len(project.get_releases())
    assert release.tag_name == increased_release
    assert release.title == "test"
    assert release.body == "testing release"
    assert count_before + 1 == count_after


@record_httpx
def test_edit_release(project):
    release = project.get_release(tag_name="0.1.0")
    origin_name = release.title
    origin_message = release.body

    release.edit_release(
        name=f"{origin_name}-changed",
        message=f"{origin_message}-changed",
    )
    assert release.title == f"{origin_name}-changed"
    assert release.body == f"{origin_message}-changed"


@record_httpx
def test_latest_release(project):
    # check the latest release of the repo and fix version and string in the body
    last_version = "0.2.11"
    release = project.get_latest_release()
    assert release.tag_name == last_version


@record_httpx
def test_latest_release_doesnt_exist(hello_world_project):
    assert hello_world_project.get_latest_release() is None



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from requre.helpers import record_httpx

from ogr.abstract import AccessLevel
from ogr.exceptions import ForgejoAPIException


@record_httpx()
def test_get_file_content(project):
    file = project.get_file_content(
        path="README.md",
        ref="b8e18207cfdad954f1b3a96db34d0706b272e6cf",
    )
    assert (
        file == "# ogr-tests\n\nTesting repository for python-ogr package. | "
        "https://github.com/packit-service/ogr\n\ntest1\ntest2\n"
    )


@record_httpx()
def test_get_file_content_resolve_dot(project):
    file = project.get_file_content(
        path="./README.md",
        ref="b8e18207cfdad954f1b3a96db34d0706b272e6cf",
    )
    assert (
        file == "# ogr-tests\n\nTesting repository for python-ogr package. | "
        "https://github.com/packit-service/ogr\n\ntest1\ntest2\n"
    )


@record_httpx()
def test_branches(project):
    branches = project.get_branches()
    assert branches
    assert "master" in set(branches)


@record_httpx()
def test_commits(project):
    commits = list(project.get_commits())
    assert commits
    assert commits[-1] == "24c86d0704694f686329b2ea636c5b7522cfdc40"


@record_httpx()
def test_get_file(project):
    file_content = project.get_file_content("README.md")
    assert file_content
    assert "Testing repository for python-ogr package." in file_content


@pytest.mark.skip(
    reason="Broken API spec resulting in failing validation in pyforgejo (#11)",
)
@record_httpx()
def test_get_files(hello_world_project):
    files = list(hello_world_project.get_files())
    assert files
    assert len(files) >= 1
    assert "README.md" in files

    files = list(hello_world_project.get_files(filter_regex=".*.md"))
    assert files
    assert len(files) >= 1
    assert "README.md" in files


@record_httpx()
def test_nonexisting_file(project):
    with pytest.raises(FileNotFoundError):
        project.get_file_content(".blablabla_nonexisting_file")


@record_httpx()
def test_get_description(project):
    description = project.get_description()
    assert description
    assert description.startswith("Testing repository for python-ogr package.")


@record_httpx()
def test_description_property(project):
    description = project.description
    assert (
        description == "Testing repository for python-ogr package.  |"
        "  https://github.com/packit-service/ogr"
    )


@record_httpx()
def test_description_setter(project):
    old_description = project.description
    assert (
        old_description == "Testing repository for python-ogr package.  |"
        "  https://github.com/packit-service/ogr"
    )

    project.description = "Different description"
    # have to refresh the cached ‹forgejo_repo› after changing the description
    del project.forgejo_repo

    assert project.description == "Different description"

    project.description = old_description
    # have to refresh the cached ‹forgejo_repo› after changing the description
    del project.forgejo_repo

    assert (
        project.description == "Testing repository for python-ogr package.  |"
        "  https://github.com/packit-service/ogr"
    )


@record_httpx()
def test_get_git_urls(project):
    urls = project.get_git_urls()
    assert urls
    assert len(urls) == 2
    assert "git" in urls
    assert "ssh" in urls
    assert urls["git"] == "https://v10.next.forgejo.org/packit-validator/ogr-tests.git"
    assert urls["ssh"].endswith(
        "git@v10.next.forgejo.org:2100/packit-validator/ogr-tests.git",
    )


@record_httpx()
def test_get_sha_from_branch(project):
    commit_sha = project.get_sha_from_branch("change")
    assert commit_sha
    assert commit_sha.startswith("d490ec67")


@record_httpx()
def test_get_sha_from_branch_non_existing(project):
    commit_sha = project.get_sha_from_branch("non-existing")
    assert commit_sha is None


@record_httpx()
def test_get_sha_from_tag(project):
    assert (
        project.get_sha_from_tag("0.1.0") == "24c86d0704694f686329b2ea636c5b7522cfdc40"
    )
    with pytest.raises(ForgejoAPIException) as ex:
        project.get_sha_from_tag("future")
    assert "404" in str(ex.value)


@record_httpx()
def test_parent_project(project):
    assert project.get_fork().parent.namespace == "packit-validator"
    assert project.get_fork().parent.repo == "ogr-tests"


@record_httpx()
def test_get_web_url(project):
    url = project.get_web_url()
    assert url == "https://v10.next.forgejo.org/packit-validator/ogr-tests"


@record_httpx()
def test_full_repo_name(project):
    assert project.full_repo_name == "packit-validator/ogr-tests"


@record_httpx()
def test_project_exists(project):
    assert project.exists()


@record_httpx()
def test_project_not_exists(service):
    assert not service.get_project(
        repo="some-non-existing-repo",
        namespace="some-none-existing-namespace",
    ).exists()


@record_httpx()
def test_is_private(service):
    # when regenerating this test with your forgejo token, use your own private repository
    private_project = service.get_project(
        namespace=service.user.get_username(),
        repo="private",
    )
    assert private_project.is_private()


@record_httpx()
def test_is_not_private(project):
    assert not project.is_private()


@record_httpx()
def test_delete(service):
    project = service.get_project(
        repo="delete-project",
        namespace=service.user.get_username(),
    )
    project.delete()


@record_httpx()
def test_has_issues(service, project):
    assert project.has_issues
    assert not service.get_project(
        namespace="packit-validator",
        repo="test",
    ).has_issues


@record_httpx()
def test_get_owners(project):
    owners = project.get_owners()
    assert owners == ["packit-validator"]


@record_httpx()
def test_get_contributors(project):
    users = project.get_contributors()
    assert users == {"mfocko", "packit-validator"}


@record_httpx()
def test_issue_permissions(project):
    users = project.who_can_close_issue()
    assert "mfocko" in users


@record_httpx()
def test_pr_permissions(project):
    users = project.who_can_merge_pr()
    assert "mfocko" in users
    assert not project.can_merge_pr("lbarcziova")


@record_httpx()
def test_get_users_with_given_access(project):
    maintainers = project.get_users_with_given_access([AccessLevel.maintain])
    assert "packit-validator" in maintainers

    admins = project.get_users_with_given_access([AccessLevel.admin])
    assert "mfocko" in admins


@record_httpx()
def test_add_remove_user(project):
    project.add_user("lbarcziova", AccessLevel.push)
    project.remove_user("lbarcziova")
